from __future__ import annotations

import copy
import warnings
from typing import TypeVar, TYPE_CHECKING, Sequence

import numpy as np
from numpy.typing import ArrayLike

from metalpy.aero.utils.line_analysis import split_lines
from metalpy.carto.coords import Coordinates
from metalpy.utils import array_ops
from metalpy.utils.array_ops import get_column
from metalpy.utils.matplotlib import check_axis
from metalpy.utils.pandas import drop_by_indices
from metalpy.utils.type import is_numeric_array, notify_package

if TYPE_CHECKING:
    import pandas as pd
    from .aerial_survey_lines import AerialSurveyLines


TAerialSurvey = TypeVar('TAerialSurvey', bound='AerialSurvey')


class AerialSurvey:
    def __init__(self, position, data=None):
        """构造航空探测数据对象

        Parameters
        ----------
        position
            坐标信息
        data
            附属的数据
        """
        self.position = position
        self.data = data

    @classmethod
    def from_array(
            cls: type[TAerialSurvey],
            df: ArrayLike | 'pd.DataFrame',
            pos_cols: Sequence[str | int] = None,
            force_lon_lat=False
    ) -> TAerialSurvey:
        if pos_cols is None:
            pos_cols = _infer_position_columns(df)
        else:
            assert len(pos_cols) == 2, '`pos_cols` should clarify exactly two columns to be horizontal position.'

        position = array_ops.get_column(df, pos_cols)

        if _guess_is_lon_lat_position(pos_cols, position) and not force_lon_lat:
            warnings.warn('Position based on longitude-latitude coordinates may interfere the line analysis process.'
                          ' Transformed to UTM local system with unit `meter`.')
            position = _wrap_lon_lat_to_utm(position=position)
            data = df
        else:
            position = array_ops.get_column(df, pos_cols)
            data = array_ops.get_columns_except(df, pos_cols)

        return cls(position, data)

    @staticmethod
    def from_photo_exifs(paths):
        import pandas as pd

        from metalpy.aero.utils.exif import load_gps_info_from_exifs

        gps = load_gps_info_from_exifs(paths)
        pos = gps[['Longitude', 'Latitude']]

        utm = _wrap_lon_lat_to_utm(position=pos)
        gps = pd.concat([utm, gps], axis=1)

        return AerialSurvey.from_array(gps)

    @classmethod
    def concat(cls: type[TAerialSurvey], lines) -> TAerialSurvey:
        pos = array_ops.concat([line.position for line in lines])
        data = [line.data for line in lines]
        if any((d is None for d in data)):
            data = None
        else:
            data = array_ops.concat(data)

        return cls(pos, data)

    @property
    def bounds(self):
        return Coordinates(self.position).bounds

    @property
    def center(self):
        return self.bounds.center.tolist()

    @property
    def length(self):
        dist = self.point_distances
        dist = dist[dist < 50]
        return dist.sum()

    @property
    def point_distances(self):
        diffs = np.diff(self.position, axis=0)
        dist = np.linalg.norm(diffs, axis=1)
        return dist

    @property
    def area(self):
        try:
            from cv2 import minAreaRect as min_area_rect
        except ImportError:
            notify_package(
                pkg_name='cv2',
                reason='`cv2.minAreaRect` is needed for more precise area computation.',
                install='pip install opencv-python'
            )
            from metalpy.utils.algos import min_area_rect_pca as min_area_rect

        rect_center, rect_size, rect_rotation = min_area_rect(
            self.position.to_numpy().astype(np.float32)
        )

        origin_area = np.prod(self.bounds.extent)
        oriented_area = np.prod(rect_size)

        return min(origin_area, oriented_area)

    @property
    def x(self):
        return get_column(self.position, 0)

    @property
    def y(self):
        return get_column(self.position, 1)

    def slice_data(self, slice0: slice | tuple, *slices: slice | tuple):
        if self.data is None:
            return None
        else:
            slices = [slice0, *slices]
            ret = []
            for s in slices:
                if not isinstance(s, slice):
                    s = slice(*s)
                ret.append(self.data[s])

            return array_ops.concat(ret)

    def slice_pos(self, slice0: slice | tuple, *slices: slice | tuple):
        slices = [slice0, *slices]
        ret = []
        for s in slices:
            if not isinstance(s, slice):
                s = slice(*s)
            ret.append(self.position[s])

        return array_ops.concat(ret)

    def slice_line(self, slice0: slice | tuple, *slices: slice | tuple):
        from . import AerialSurveyLine

        return AerialSurveyLine(
            self.slice_pos(slice0, *slices),
            self.slice_data(slice0, *slices),
        )

    def slice_lines(self, segments) -> 'AerialSurveyLines':
        from . import AerialSurveyLines

        return AerialSurveyLines([self.slice_line(seg) for seg in segments])

    def extract_lines(
            self,
            auto_clear=True,
            window_size=None,
            window_length=10,
            stable_window_size=None,
            stable_window_length=50,
            stable_direction_tol: float = 3,
            merge=True
    ) -> 'AerialSurveyLines':
        """分割提取测区中的测线

        Parameters
        ----------
        auto_clear
            是否自动清理航向角与主航向角差距较大的航段（通常为辅助航段，例如起降的航段）。
            如果包含多方向飞行的航线则不应该启用此选项。
        window_size, window_length
            用于分析航向的窗口测点数。应当大于飞机产生明显位移的窗口大小。
            过小时可能会导致测线划分对飞机飞行扰动敏感产生错误划分，过大时可能导致转弯等非平稳飞行测点也被容纳进测线。
        stable_window_size
            用于确定为航线的窗口测点数。应当大于转弯等非平稳飞行过程的窗口大小。
            过小时可能导致转弯等非平稳测点被统计为测线，过大时可能导致某些较短测线被意外移除。
        window_length, stable_window_length
            对应的窗口长度。若不指定对应size而指定length，则自动统计测点间距然后通过length求取窗口测点数。
        stable_direction_tol
            用于判断是否归属于相同测线的航向容差，单位（度）
        merge
            指示是否在自动清理辅助航段后（`auto_clear=True`）再次合并相邻航线。
            通常用于分多次完成的大测区测量，中间涉及多次起飞降落的辅助航段，移除后重新合并为完整测区的航线。

        Returns
        -------
        lines
            分析提取的各条测线

        Notes
        -----
        航段整体上可以分为三类：工作区间，辅助区间和非平稳区间
            1. 工作区间为正常平稳飞行，测量数据使用的飞行区间

            2. 辅助区间为平稳飞行的区间，但并不进行数据测量，例如飞机起飞降落、转向等动作

            3. 非平稳区间，飞行状态切换之间的状态，例如高速转向、悬停等状态下，会被判定为非平稳区间

        `extract_lines` 实现了排除 `非平稳区间` ，并提取切分 `工作区间` 和 `辅助区间` ，
        然后通过 `remove_auxiliary_lines` 判断并移除 `辅助区间` 。
        """
        xs, ys = np.asarray(self.position).T
        segments = split_lines(
            xs, ys,
            window_size=window_size,
            window_length=window_length,
            stable_window_size=stable_window_size,
            stable_window_length=stable_window_length,
            stable_direction_tol=stable_direction_tol
        )

        assert len(segments) > 0, 'Failed to pre-analyze flight lines by heading direction.'

        lines = self.slice_lines(segments)
        if auto_clear:
            lines.remove_auxiliary(inplace=True)
        if merge:
            lines.merge_neighbors(inplace=True)

        return lines

    def trim_auxiliary(
            self,
            keep_takeoff_and_landing=False,
            interval_threshold=None,
            ref_lines: 'AerialSurveyLines' | None = None,
            aux_lines: 'AerialSurveyLines' | None = None,
            inplace=False
    ) -> 'AerialSurvey':
        """基于`extract_lines`的测线划分结果，从原始测线数据中清理掉辅助轨迹，从而合并为一个单航次的数据

        例如对于较大测区，需要分多航次进行测量时，清除掉中途的多次起飞降落的轨迹，结果可以视为一个航次内完成采集的数据

        区别于`extract_lines`，转向过程等常规航次应有的轨迹会保留

        Parameters
        ----------
        keep_takeoff_and_landing
            指示是否保留起飞和降落轨迹，为False则只保留工作状态下的航线轨迹
        interval_threshold
            跳变点判定阈值，若不指定，则取为航线间距
        ref_lines
            用作参考的工作航线划分，若不指定，则使用默认参数进行航线划分
        aux_lines
            用作参考的辅助航线划分，若不指定，则使用默认参数进行航线划分
        inplace
            指示是否就地操作，默认为False，即返回修改后的实例，若否，则直接修改当前实例

        Returns
        -------
        trimmed
            移除掉辅助轨迹后的Survey实例

        Notes
        -----
        如果外部给定`ref_lines`参数，则注意需要指定`auto_clear`和`merge`为False来保留识别到的辅助航段信息

        >>> survey = AerialSurvey.from_array(...)
        >>> lines = survey.extract_lines(
        >>>     auto_clear=False,
        >>>     merge=False
        >>>     stable_direction_tol=2.5
        >>> )
        >>> survey.trim_auxiliary(ref_lines=lines)
        """
        import pandas as pd

        assert hasattr(self.data, 'index'), \
            ('`remove_auxiliary` requires indexed arrays,'
             ' try using pd.DataFrame as array type.'
             '\n'
             '    AerialSurvey.from_array(pd.DataFrame(your_data))'
             '\n')

        if ref_lines is None:
            ref_lines = self.extract_lines(
                auto_clear=False,
                merge=False,
            )
            aux_lines = ref_lines.pop_auxiliary()
            ref_lines.merge_neighbors(inplace=True)
        else:
            ref_lines = copy.deepcopy(ref_lines)

        specs = ref_lines.get_line_specs()
        line_interval = np.max([s.line_interval for s in specs if not np.isnan(s.line_interval)])
        aux_indices = []

        if aux_lines is not None:
            # 排除已经判定为辅助航线的部分
            # 注意，拐弯部分也有可能会被判定为辅助航线，因此判断和`line_interval`航线间隔相差过大的为辅助航线而非拐弯
            for line in aux_lines:
                if max(line.position.index) <= min(ref_lines[0].position.index):
                    continue  # 第一次起飞轨迹会在后方单独讨论
                if min(line.position.index) >= max(ref_lines[-1].position.index):
                    continue  # 最后一次降落轨迹会在后方单独讨论
                if not (0.5 <= line.length / line_interval <= 2):
                    aux_indices.append(line.position.index)

        if not keep_takeoff_and_landing:
            # 第一条测线之前和最后一条测线之后，分别对于第一次起飞和最后一次降落的轨迹
            aux_indices.extend([
                pd.RangeIndex(0, min(ref_lines[0].position.index)),
                pd.RangeIndex(max(ref_lines[-1].position.index), max(self.data.index) + 1),
            ])

        for line in ref_lines:
            # 所有已经判定为工作航线的区间，如果并非连续的数据点，则中间间断的部分也可能是辅助航线
            ref = line.position.index
            if not isinstance(ref, pd.RangeIndex):
                interval_pos = np.where(np.diff(ref) > 1)[0]
                for ip in interval_pos:
                    aux_indices.append(pd.RangeIndex(ref[ip] + 1, ref[ip + 1]))

        dropped_pos = drop_by_indices(self.position, aux_indices)
        dropped_flight = AerialSurvey.from_array(dropped_pos)
        pos = dropped_flight.position

        # 检查移除相关航线后导致留下的间断航段
        dist = np.linalg.norm(np.diff(pos, axis=0), axis=1)
        normal = np.percentile(dist, 50)
        if interval_threshold is None:
            # 不给定间断阈值的话，采用航线间距作为判定阈值
            interval_threshold = line_interval

        threshold_pos = np.where(dist - normal > interval_threshold)[0]
        threshold_pos = np.r_[-1, threshold_pos, len(pos) - 1]
        windows = np.lib.stride_tricks.sliding_window_view(threshold_pos, 2) + 1
        segments = AerialSurvey(pos).slice_lines(windows)

        # 依据间断距离分割的各个大航段，如果存在工作航线属于该航段，则保留
        # 若不存在，则可能为之前未清理干净的辅助轨迹，需要移除
        idling_indices = []
        i = 0
        for segment in segments:
            contains_line = False

            seg_index = segment.position.index
            smin, smax = min(seg_index), max(seg_index)

            while i < len(ref_lines):
                ref = ref_lines[i]
                ref_ind = ref.position.index
                if min(ref_ind) < smin:
                    # 工作航线在航段之前
                    i += 1
                elif max(ref_ind) < smax:
                    # 工作航线包含在航段之中
                    contains_line = True
                    i += 1
                    break
                else:
                    # 工作航线在航段之后
                    break

            if not contains_line:
                idling_indices.append(seg_index)

        dropped_pos = drop_by_indices(dropped_pos, idling_indices)
        if self.data is not None:
            dropped_data = drop_by_indices(self.data, aux_indices + idling_indices)
        else:
            dropped_data = None

        if inplace:
            self.position = dropped_pos
            self.data = dropped_data
            ret = self
        else:
            ret = AerialSurvey(dropped_pos, dropped_data)

        return ret

    def plot(self, ax=None, show=True, **kwargs):
        with check_axis(ax, show=show, figsize=(5, 5)) as ax:
            ax.plot(*np.asarray(self.position).T, **kwargs)

    def scatter(self, ax=None, show=True, color=True, **kwargs):
        with check_axis(ax, show=show, figsize=(5, 5)) as ax:
            if 's' not in kwargs:
                kwargs['s'] = 0.05  # 默认的点尺寸太大，会互相重叠

            if color and 'c' not in kwargs:
                kwargs['c'] = range(len(self.position))

            ax.scatter(*np.asarray(self.position).T, **kwargs)

    def to_polydata(
            self,
            *,
            z=None,
            data_col: str | int = None,
            points_only=False,
            color=True
    ):
        """导出航空数据为PyVista模型

        Parameters
        ----------
        z
            指定z值，或者通过字符串指定z值所在列。
            如果给定为单个数值类型，则同时用作所有数据点的高度值。
        data_col
            指定作为z值的数据列，支持索引或列名
        points_only
            指示返回值是否只包含点坐标模型，不包含连接线
        color
            指示返回值是否包含顺序信息，默认为True，则PointData内绑定数据下标，渲染时自动着色

        Returns
        -------
        ret
            生成的PyVista模型
        """
        import pyvista as pv

        if isinstance(z, str):
            data_col = z
            z = None

        if data_col is not None:
            z = array_ops.get_column(self.data, data_col)
        else:
            if z is None:
                z = 0
            if not is_numeric_array(z):
                z = np.full(len(self.position), z)

        pos = np.c_[self.position, z]
        n_pts = pos.shape[0]

        ret = pv.PolyData(pos)

        if not points_only:
            ret.lines = np.r_[n_pts, range(n_pts)]

        if color:
            ret.point_data['index'] = range(n_pts)

        return ret

    def to_planar(self):
        from metalpy.aero.routes import FlightPlanar
        planar = FlightPlanar(np.percentile(self.point_distances, 50))
        planar.add_path(self.position)

        return planar

    def __len__(self):
        return len(self.position)


POSITION_KEYWORDS = [
    ['X', 'Y'],
    ['Longitude', 'Latitude'],
    ['Long', 'Lat'],
    ['Lon', 'Lat'],
]
POSITION_KEYWORDS = [
    (proc(xk), proc(yk))
    for xk, yk in POSITION_KEYWORDS
    for proc in [
        lambda k: k.capitalize(),
        lambda k: k.upper(),
        lambda k: k.lower(),
    ]
][1:]


def _infer_position_columns(arr):
    if hasattr(arr, 'keys'):
        def _infer():
            keys: set[str] = set(arr.keys())
            for xkey, ykey in POSITION_KEYWORDS:
                if xkey in keys and ykey in keys:
                    return [xkey, ykey]

                for possible_xkey in keys:
                    segs = possible_xkey.split(xkey)
                    n_segs = len(segs)
                    if n_segs <= 1:
                        continue

                    for i in range(1, n_segs):
                        left = xkey.join(segs[:i])
                        right = xkey.join(segs[i+1:])
                        inferred_ykey = f'{left}{ykey}{right}'

                        if inferred_ykey in keys:
                            return [possible_xkey, inferred_ykey]

            return None

        keys_found = _infer()
    else:
        keys_found = None

    if keys_found is None:
        keys_found = [0, 1]

    return keys_found


def _guess_is_lon_lat_position(cols: Sequence[str | int], position):
    if isinstance(cols[0], str) and 'lon' in cols[0].lower() and 'lat' in cols[1].lower():
        return True
    else:
        x, y = get_column(position, 0), get_column(position, 1)
        if (
                all(abs(x) <= 180)  # 经度范围
                and all(abs(y) <= 90)  # 纬度范围
                and all(np.std(x) < 10)  # 如果单位为m，则主要区域范围不应小于10m，此时应为误用经纬度
        ):
            return True
        else:
            return False


WrapLonLatToUTM = array_ops.OperatorDispatcher('WrapLonLatToUTM')


def _wrap_lon_lat_to_utm(position):
    return WrapLonLatToUTM.dispatch(type(position))(position)


@WrapLonLatToUTM.register(array_ops.ArrayType.numpy)
def _wrap_lon_lat_to_utm_np(position):
    coords = Coordinates(position, crs=Coordinates.WGS_84).warp(query=Coordinates.SearchUTM)
    return coords.to_numpy()


@WrapLonLatToUTM.register(array_ops.ArrayType.pandas)
def _wrap_lon_lat_to_utm_pd(position):
    import pandas as pd
    coords = Coordinates(position, crs=Coordinates.WGS_84).warp(query=Coordinates.SearchUTM)
    return pd.DataFrame(coords, columns=['UTMX', 'UTMY'], index=position.index)



from __future__ import annotations

import abc
import warnings
from typing import Iterable, TYPE_CHECKING, Union

import numpy as np

from metalpy.utils.batch import Batch
from metalpy.utils.rand import check_random_state
from .worker import Worker

if TYPE_CHECKING:
    from .executor import Executor

WeightsType = Union[Iterable[Union[Worker, float]], 'Executor']
ShuffleChoice = Union[bool, int, np.random.RandomState, np.ndarray]


class TaskAllocator(abc.ABC):
    @abc.abstractmethod
    def slice(self, n_slices=1):
        pass

    @abc.abstractmethod
    def assign(self, worker_or_weight: Worker | float):
        pass

    @abc.abstractmethod
    def iter(self, var_inp: WeightsType):
        pass

    @property
    @abc.abstractmethod
    def finished(self):
        pass

    @abc.abstractmethod
    def inverse_shuffle(self, *data):
        pass

    @abc.abstractmethod
    def reassemble(self, *data, reverse_shuffle=True):
        pass

    @abc.abstractmethod
    def __getitem__(self, item):
        pass


class SingleTaskAllocator(TaskAllocator):
    def __init__(self, data, *, total=None, shuffle: ShuffleChoice = False):
        self.data = data
        if total is None:
            total = len(data)
        self.total = total
        self.base_index = 0

        if shuffle:
            if isinstance(shuffle, np.ndarray):
                assert len(shuffle) == len(data), \
                    '`shuffle` should be random mask when provided with ndarray.'
                self.shuffle_mask = shuffle
            else:
                self.shuffle_mask = np.arange(len(data))
                if shuffle is True:
                    shuffle = 0x0
                shuffle = check_random_state(shuffle)
                shuffle.shuffle(self.shuffle_mask)
        else:
            self.shuffle_mask = None

    def slice(self, n_slices=1):
        end = self.base_index + int(np.ceil(len(self.data) * n_slices / self.total))
        if self.shuffle_mask is not None:
            mask = self.shuffle_mask[self.base_index:end]
        else:
            mask = slice(self.base_index, end)

        ret = self.data[mask]
        self.base_index = end
        return ret

    def assign(self, worker_or_weight: Worker | float):
        if isinstance(worker_or_weight, Worker):
            worker_or_weight = worker_or_weight.get_weight()

        return self.slice(worker_or_weight)

    def iter(self, var_inp: WeightsType):
        for w in extract_weights(var_inp):
            yield self.assign(w)

    @property
    def finished(self):
        return self.base_index >= len(self.data)

    def inverse_shuffle(self, *data):
        """从随机打乱的数据中恢复原索引

        Parameters
        ----------
        data
            乱序的数据，可能是分派任务后产出的数据

        Returns
        -------
        ordered_data
            还原为原顺序的数据
        """
        assert len(data) == 1, f'`{SingleTaskAllocator.__name__}` accepts only one data.'
        data = data[0]

        if self.shuffle_mask is None:
            return data
        else:
            inverse_mask = self.shuffle_mask.copy()
            inverse_mask[self.shuffle_mask] = np.arange(len(self.shuffle_mask))
            return np.asarray(data)[inverse_mask]

    def reassemble(self, *data, reverse_shuffle=True):
        """从被分组的数据中重新组装，并恢复随机打乱过程

        Parameters
        ----------
        data
            乱序的数据列表，可能是分派任务后产出的数据
        reverse_shuffle
            指示是否将拼装后的结果从乱序中复原

        Returns
        -------
        reassembled_data
            组装后的数据
        """
        assert len(data) == 1, f'`{SingleTaskAllocator.__name__}` accepts only one data.'
        data = data[0]

        data = np.concatenate(data, axis=0)

        if reverse_shuffle:
            data = self.inverse_shuffle(data)

        return data

    def __getitem__(self, item):
        if item > 0:
            warnings.warn(f'Trying to access single allocator with index {item}, got self anyway.')
        return self


class MultiTaskAllocator(TaskAllocator):
    def __init__(self, *data_list, total=None, shuffle: ShuffleChoice = False):
        allocators = [SingleTaskAllocator(data, total=total, shuffle=shuffle) for data in data_list]
        self.allocators = Batch.of(allocators)

    def slice(self, n_slices=1):
        return list(self.allocators.slice(n_slices))

    def assign(self, worker_or_weight: Worker | float):
        return list(self.allocators.assign(worker_or_weight))

    def iter(self, var_inp: WeightsType):
        yield from zip(*self.allocators.iter(var_inp))

    @property
    def finished(self):
        return any(self.allocators.finished)

    def shuffle_masks(self):
        return [
            a.shuffle_mask for a in self.allocators
        ]

    def inverse_shuffle(self, *data_list):
        return [
            a.inverse_shuffle(d) for d, a in zip(data_list, self.allocators)
        ]

    def reassemble(self, *data_list, reverse_shuffle=True):
        return [
            a.reassemble(d, reverse_shuffle=reverse_shuffle)
            for d, a in zip(data_list, self.allocators)
        ]

    def __getitem__(self, item):
        return self.allocators.batch_items[item]


class BoundTaskAllocator(TaskAllocator):
    def __init__(self, allocator: TaskAllocator, weights: WeightsType):
        self.allocator = allocator
        self.weights = extract_weights(weights)

    @staticmethod
    def arrange(data, *, weights, shuffle: ShuffleChoice = False):
        weights = extract_weights(weights)
        return BoundTaskAllocator(
            SingleTaskAllocator(data, total=sum(weights), shuffle=shuffle),
            weights=weights
        )

    @staticmethod
    def arrange_many(*data_list, weights, shuffle: ShuffleChoice = False):
        weights = extract_weights(weights)
        return BoundTaskAllocator(
            MultiTaskAllocator(*data_list, total=sum(weights), shuffle=shuffle),
            weights=weights
        )

    def slice(self, n_slices=1):
        return self.allocator.slice(n_slices)

    def assign(self, worker: Worker | float):
        return self.allocator.assign(worker)

    def iter(self, var_inp: WeightsType):
        yield from self.allocator.iter(var_inp)

    @property
    def finished(self):
        return self.allocator.finished

    def inverse_shuffle(self, *data):
        return self.allocator.inverse_shuffle(*data)

    def reassemble(self, *data_list, reverse_shuffle=True):
        return self.allocator.reassemble(*data_list, reverse_shuffle=reverse_shuffle)

    def __iter__(self):
        yield from self.allocator.iter(self.weights)

    def __getitem__(self, item):
        return self.allocator[item]


def extract_weights(weights: WeightsType):
    from .executor import Executor

    if isinstance(weights, Executor):
        weights = weights.get_workers()

    if isinstance(weights, Iterable):
        return [_extract_weight(w) for w in weights]

    return _extract_weight(weights)


def _extract_weight(weight: WeightsType):
    if isinstance(weight, Worker):
        return weight.get_weight()
    else:
        return weight  # 应该得是float，大概



import numpy as np
from numpy.testing import assert_almost_equal

from metalpy.utils.sensor_array import get_grids_ex


def test_get_grids_ex():
    arrs = [
        # xs, ys, zs = [0, 1, 2, 3], [0, 1, 2, 3], [0]
        get_grids_ex(xs=np.arange(4), ys=np.arange(4), zs=0).pts,
        get_grids_ex(cell_width=[1, 1], nz=1, origin=[0, 0, 0], end=[3, 3, 3]).pts,
        get_grids_ex(n=4, nz=1, origin=[0, 0, 0], end=[3, 3, 3]).pts,
        get_grids_ex(n=4, nz=1, cell_width=[1, 1], origin=[0, 0, 0]).pts,
        get_grids_ex(n=4, nz=1, cell_width=[1, 1], end=[3, 3, 0]).pts,
    ]

    for arr in arrs:
        assert_almost_equal(arrs[0], arr)

    arr2 = get_grids_ex(n=4, cell_width=[1, 1], end=[3, 3, 0]).pts
    assert arr2.ndim == 2



from __future__ import annotations

import datetime
import json
import re
import warnings
from bisect import bisect, bisect_right, bisect_left
from collections import namedtuple, OrderedDict

import numpy as np
import requests

from metalpy.utils.bounds import Bounds
from metalpy.utils.file import file_cached
from .simple_tile_map_source import SimpleTileMapSource

# 版本年份，修订号
# 注：版本年份不一定和title中的发布日期一致，特例为2017r21对应2018-01-18。
Revision = namedtuple('Revision', ['year', 'rev'])
# 拍摄日期，影像来源，影像来源描述（拍摄卫星），分辨率，定位精度（与真实地理位置的偏差）
SourceMetaData = namedtuple('SourceMetaData', ['capture_date', 'source', 'desc', 'resolution', 'accuracy'])


class WaybackSource(SimpleTileMapSource):
    Sources: list[WaybackSource] = []
    SourcesByRnum: dict[int, WaybackSource] = {}

    Baseurl = 'https://wayback.maptiles.arcgis.com/arcgis/rest/services/World_Imagery'
    TileUrl = Baseurl + '/WMTS/1.0.0/default028mm/MapServer/tile/{rnum}/{level}/{row}/{col}'  # 默认瓦片地图信息地址
    TileInfoUrl = Baseurl + '/MapServer/tilemap/{rnum}/{level}/{row}/{col}'  # 默认瓦片地图信息地址

    MetaDataBaseurl = 'https://metadata.maptiles.arcgis.com/arcgis/rest/services'
    MetaDataUrl = MetaDataBaseurl + '/World_Imagery_Metadata_{year}_r{rev}/MapServer'  # 默认瓦片地图元数据地址

    # TODO: 读取官方WMTS服务
    #  https://wayback.maptiles.arcgis.com/arcgis/rest/services/world_imagery/wmts/WMTSCapabilities.xml
    ConfigUrl = 'https://s3-us-west-2.amazonaws.com/config.maptiles.arcgis.com/waybackconfig.json'  # Wayback配置地址

    MAX_ZOOM = 23
    MIN_ZOOM = 10

    def __init__(self, title, date, rev: Revision, rnum=None, tile_url=None, metadata_url=None):
        """Wayback (esri) 地图数据源

        References
        ----------
        World Imagery Wayback: https://livingatlas.arcgis.com/wayback/
        """
        self.title = title
        self.date = date
        self.revision = rev
        self.release_num = rnum

        if tile_url is None:
            tile_url = WaybackSource.format_tile_url(rnum)

        if metadata_url is None:
            metadata_url = WaybackSource.format_metadata_url(self.revision)

        self.tile_url = tile_url
        self.metadata_url = metadata_url

        super().__init__(
            tile_url
        )

    @property
    def year(self):
        return self.revision.year

    @property
    def rev(self):
        return self.revision.rev

    def query_metadata(self, mercator_bounds: Bounds, level, session=None):
        """查询区域内卫星地图的元数据，包括影像采集时间，影像源，采集卫星，分辨率，精度等。

        Parameters
        ----------
        mercator_bounds
            区域边界
        level
            瓦片地图等级
        session
            requests会话，若为空则直接使用request

        Returns
        -------
        metas
            元数据列表，因为区域内可能包含多个来源的数据
        """
        mercator_bounds = Bounds(mercator_bounds)
        geometry = {
            'spatialReference': {'latestWkid': 3857, 'wkid': 102100},
            'xmin': mercator_bounds.xmin,
            'xmax': mercator_bounds.xmax,
            'ymin': mercator_bounds.ymin,
            'ymax': mercator_bounds.ymax,
        }
        params = {
            'f': 'json',
            'outFields': ','.join(['SRC_DATE2', 'NICE_DESC', 'SRC_DESC', 'SRC_RES', 'SRC_ACC']),
            'geometry': json.dumps(geometry),
            'geometryType': 'esriGeometryEnvelope',
            'spatialRel': 'esriSpatialRelIntersects',
            'returnGeometry': 'false'
        }
        layer_id = WaybackSource.to_metadata_layer_id(level)

        if session is None:
            session = requests

        resp = session.get(self.metadata_url + f'/{layer_id}/query', params=params)
        metas = resp.json()

        metas = [WaybackSource.to_metadata(meta) for meta in metas['features']]

        return metas

    @staticmethod
    def query_tile_changes(mercator_bounds, level, session=None, region_samples: int | float | None = 0.2):
        """查询区域内瓦片存在变更的时间点（仍可能存在重复）

        Parameters
        ----------
        mercator_bounds
            区域边界
        level
            瓦片地图等级
        session
            requests会话，若为空则内部创建
        region_samples
            区域采样，
            如果为整数则为固定个数采样，
            如果为小数则按比例采样，
            如果为空则不进行采样

        Returns
        -------
        wayback_sources
            区域内瓦片发生变更的时间点的地图源

        TODO: 实现并行查询
        TODO: 实现进一步去重
          （https://github.com/vannizhang/wayback/blob/master/src/core/WaybackManager/ChangeDetector.ts#L221）
        """
        from metalpy.carto.basemap.constants import WebMercator
        _1day = datetime.timedelta(days=1)
        sources = OrderedDict()
        latest_rnum = WaybackSource.get_sources()[-1].release_num

        if session is None:
            session = requests.session()

        tiles = list(WebMercator.iter_tiles(mercator_bounds, level))
        if isinstance(region_samples, float):
            assert 0 < region_samples < 1, 'Proportional region samples must be within (0.0, 1.0).'
            region_samples = int(len(tiles) * region_samples)

        if isinstance(region_samples, int):
            assert region_samples > 1, 'Fixed region samples must be larger than 0.'
            if region_samples < len(tiles):
                sample_indices = np.linspace(0, len(tiles) - 1, region_samples).astype(int)
                tiles = [tiles[i] for i in sample_indices]

        for col, row in tiles:
            last_release_num = latest_rnum

            while last_release_num is not None:
                url = WaybackSource.format_tile_info_url(last_release_num, level, col, row)
                info = session.get(url).json()
                if 'select' in info and len(info['select']) > 0:
                    last_release_num = info['select'][0]
                    release = sources[last_release_num] = WaybackSource.find_by_release_num(last_release_num)
                else:
                    release = WaybackSource.find_by_release_num(last_release_num)

                if 'data' in info and info['data'][0]:
                    last_release = WaybackSource.find_latest_before(release.date - _1day)
                    if last_release is not None:
                        last_release_num = last_release.release_num
                    else:
                        last_release_num = None
                else:
                    last_release_num = None

        ret = list(sources.values())
        ret.sort()

        return ret

    @staticmethod
    def find_by_rev(year: int | None = None, rev: int | Revision = None):
        if year is None:
            year = rev.year
            rev = rev.rev
        if not isinstance(rev, Revision):
            rev = Revision(year=year, rev=rev)
        return WaybackSource._find_by(rev)

    @staticmethod
    def find_by_date(date):
        return WaybackSource._find_by(date)

    @staticmethod
    def find_by_release_num(rnum):
        return WaybackSource.get_sources_from_rnum().get(rnum)

    @staticmethod
    def findall_by_year(year):
        sources = WaybackSource.get_sources()
        left = bisect(sources, datetime.date(year, 1, 1))
        right = bisect(sources, datetime.date(year + 1, 1, 1), lo=left)

        return sources[left:right]

    @staticmethod
    def findall_before(date: datetime.date):
        sources = WaybackSource.get_sources()
        right = bisect_left(sources, date)
        return sources[:right]

    @staticmethod
    def findall_after(date: datetime.date):
        sources = WaybackSource.get_sources()
        left = bisect_right(sources, date)
        return sources[left:]

    @staticmethod
    def find_latest_before(date: datetime.date):
        sources = WaybackSource.get_sources()
        right = bisect_right(sources, date)
        if right <= 0:
            return None
        return sources[right - 1]

    @staticmethod
    def find_earliest_after(date: datetime.date):
        sources = WaybackSource.get_sources()
        left = bisect_left(sources, date)
        if left > len(sources):
            return None
        return sources[left]

    @staticmethod
    def _find_by(date_or_rev):
        sources = WaybackSource.get_sources()
        index = bisect_left(sources, date_or_rev)
        if sources[index] != date_or_rev:
            return None
        return sources[index]

    @staticmethod
    def from_wayback_item(rnum, item_info):
        match = re.compile(r'(\d+)_r(\d+)').search(item_info['metadataLayerUrl'])
        revision = Revision(year=int(match[1]), rev=int(match[2]))

        date = re.compile(r'\d{4}-\d{2}-\d{2}').search(item_info['itemTitle'])[0]
        date = datetime.date.fromisoformat(date)

        urls = {}
        tile_url = item_info['itemURL']
        if tile_url != WaybackSource.format_tile_url(rnum):
            warnings.warn('Tile url updated!'
                          ' Please open an issue on `https://github.com/yanang007/metalpy/issues`'
                          ' to report this change.')
            urls['tile_url'] = tile_url

        metadata_url = item_info['metadataLayerUrl']
        if metadata_url != WaybackSource.format_metadata_url(revision):
            warnings.warn('Metadata url updated!'
                          ' Please open an issue on `https://github.com/yanang007/metalpy/issues`'
                          ' to report this change.')
            urls['metadata_url'] = metadata_url

        return WaybackSource(
            title=item_info['itemTitle'],
            date=date,
            rev=revision,
            rnum=int(rnum),
            **urls
        )

    @staticmethod
    def get_sources(force_update=False):
        if len(WaybackSource.Sources) == 0:
            WaybackSource.Sources = []
            configs = WaybackSource.get_config.prepare(update=force_update)()
            for rnum, info in configs.items():
                source = WaybackSource.from_wayback_item(rnum, info)
                WaybackSource.Sources.append(source)
                WaybackSource.SourcesByRnum[source.release_num] = source

        WaybackSource.Sources.sort()
        return WaybackSource.Sources

    @staticmethod
    def get_sources_from_rnum(force_update=False):
        WaybackSource.get_sources(force_update)
        return WaybackSource.SourcesByRnum

    @staticmethod
    @file_cached(ttl=datetime.timedelta(days=3))
    def get_config():
        resp = requests.get(WaybackSource.ConfigUrl)
        return resp.json()

    @staticmethod
    def format_tile_url(rnum):
        return WaybackSource.TileUrl.replace('{rnum}', str(rnum))

    @staticmethod
    def format_tile_info_url(rnum, level, col, row):
        return WaybackSource.TileInfoUrl.format(rnum=rnum, level=level, col=col, row=row)

    @staticmethod
    def format_metadata_url(rev: Revision):
        return WaybackSource.MetaDataUrl.format(year=rev.year, rev=f'{rev.rev:02}')

    @staticmethod
    def to_metadata_layer_id(level):
        layer_id = WaybackSource.MAX_ZOOM - level
        layer_id = max(layer_id, 0)
        layer_id = min(layer_id, WaybackSource.MAX_ZOOM - WaybackSource.MIN_ZOOM)

        return layer_id

    @staticmethod
    def to_metadata(meta):
        meta = meta['attributes']
        date = datetime.datetime.fromtimestamp(int(meta['SRC_DATE2']) // 1000)
        date = date.astimezone(datetime.timezone.utc)
        return SourceMetaData(
            capture_date=date,
            source=meta['NICE_DESC'],
            desc=meta['SRC_DESC'],
            resolution=meta['SRC_RES'],
            accuracy=meta['SRC_ACC'],
        )

    def __eq__(self, other):
        return self._cmp(other) == 0

    def __le__(self, other):
        return self._cmp(other) <= 0

    def __ge__(self, other):
        return self._cmp(other) >= 0

    def __lt__(self, other):
        return self._cmp(other) < 0

    def __gt__(self, other):
        return self._cmp(other) > 0

    def __ne__(self, other):
        return self._cmp(other) != 0

    def _cmp(self, other):
        if isinstance(other, WaybackSource):
            return (self.date - other.date).days
        elif isinstance(other, datetime.date):
            return (self.date - other).days
        elif isinstance(other, Revision):
            dyear = self.year - other.year
            if dyear == 0:
                return self.rev - other.rev
            else:
                return dyear
        else:
            return self.date - other



import numpy as np

from metalpy.utils.rand import check_random_state


def extract_vec(pool, random_state):
    n_edges = len(pool)
    pool = np.sort(pool)
    vec = []

    xmin, xmax = pool[0], pool[-1]

    last_left, last_right = xmin, xmin
    for i in range(1, n_edges - 1):
        x = pool[i]

        if random_state.rand() > 0.5:
            vec.append(x - last_left)
            last_left = x
        else:
            vec.append(last_right - x)
            last_right = x

    vec.append(xmax - last_left)
    vec.append(last_right - xmax)

    return vec


def gen_random_convex_polygon(n_edges, size=None, random_state=None):
    """
    https://kingins.cn/2022/02/18/%E9%9A%8F%E6%9C%BA%E5%87%B8%E5%A4%9A%E8%BE%B9%E5%BD%A2%E7%94%9F%E6%88%90%E7%AE%97%E6%B3%95/
    Valtr P . Probability thatnrandom points are in convex position[J]. 1995, 13(1):637-643.

    :param size: 随机多边形的最大尺寸
    :param n_edges: 边数
    :param random_state: 随机状态
    :return: 随机多边形的各顶点
    """

    if size is None:
        size = [1, 1]
    size = np.asarray(size)

    # np.random.Generator
    random_state = check_random_state(random_state)

    # generate two lists of random X and Y coordinates
    x_pool = random_state.rand(n_edges)
    y_pool = random_state.rand(n_edges)

    # sort
    x_pool = np.sort(x_pool)
    y_pool = np.sort(y_pool)

    # isolate the extreme points
    xmin, xmax = x_pool[0], x_pool[-1]
    ymin, ymax = y_pool[0], y_pool[-1]

    # divide the interior points into two chains and extract the vector components
    x_vec = extract_vec(x_pool, random_state)
    y_vec = extract_vec(y_pool, random_state)

    # randomly pair up the X- and Y- components
    random_state.shuffle(y_vec)

    # combine the paired up components into vectors
    vectors = [np.asarray((x, y)) for x, y in zip(x_vec, y_vec)]
    vectors.sort(key=lambda a: np.arctan2(*a))

    # lay them end to end
    point = np.asarray((0, 0))
    minPolygonX = 0.0
    minPolygonY = 0.0
    points = []

    for vector in vectors:
        points.append(point)
        point = point + vector
        minPolygonX = min(minPolygonX, point[0])
        minPolygonY = min(minPolygonY, point[1])

    points = np.vstack(points)
    # translate the polygon to the original min and max coordinates
    translation = np.asarray((xmin, ymin)) - np.asarray((minPolygonX, minPolygonY))

    result = (points + translation) * size

    return result


from typing import NamedTuple

from ._version import get_version, get_version_tuple

__version__ = get_version()
version_tuple = get_version_tuple()
full_version = __version__
base_version = NamedTuple('ShortVersion', major=int, minor=int, patch=int)(*version_tuple[:3])
short_version = '.'.join((str(i) for i in version_tuple[:3]))
is_release = len(version_tuple) == 3

post_distance = 0
revision_prefix = 'post'
if not is_release:
    for ver_desc in version_tuple[3:]:
        if ver_desc.startswith(revision_prefix):
            post_distance = int(version_tuple[3][len(revision_prefix):])
            break

del get_version, get_version_tuple, revision_prefix



from numpy.testing import assert_almost_equal

from metalpy.scab.modelling.shapes import Ellipsoid


def test_shape3d_to_polydata_with_transforms():
    # 测试transforms的等价性
    sphere = Ellipsoid(10, 8, 6)
    raw = sphere.to_local_polydata()

    delta = [5, 0, 0]
    translated = sphere.translate(*delta)
    translated_model = translated.to_polydata()
    translated_model_truth = raw.translate(delta)
    assert_almost_equal(translated_model_truth.points, translated_model.points, decimal=6)

    rotated = translated.rotate(90, 90, 0, degrees=True, seq='xyz')  # vtk使用的是外旋
    rotated_model = rotated.to_polydata()
    rotated_model_truth = translated_model_truth.rotate_x(90).rotate_y(90)
    assert_almost_equal(rotated_model.points, rotated_model_truth.points, decimal=6)



import psutil
from loky import get_reusable_executor

from .pool_executor import PoolExecutor
from .worker import Worker


class ProcessExecutor(PoolExecutor):
    def __init__(self, n_units=None):
        """构造基于loky进程池的执行器

        Parameters
        ----------
        n_units
            工作单位数，一般为物理核心数

        See Also
        --------
        :class:`ProcessPoolExecutor` : 原生进程池
        """
        if n_units is None:
            n_units = psutil.cpu_count(logical=False)

        pool = get_reusable_executor(max_workers=n_units, kill_workers=True)
        workers = [Worker(f'proc-{i}', 1) for i in range(n_units)]

        super().__init__(pool_executor=pool, workers=workers)

    def _get_queue(self):
        return self.pool._context.Manager().Queue()



from metalpy.utils.object_path import ObjectPath


def reget_object(obj):
    """用于获取被劫持后的对象
    
    由于mexin中的劫持是通过替换其上层对象的成员实现的，
    因此在劫持发生前获取的引用不会被劫持影响。

    Parameters
    ----------
    obj
        劫持前获取的对象

    Returns
    -------
    ret
        同路径下的对象，若没有劫持行为，则返回值和输入指向同一个对象
    """
    return ObjectPath.of(obj).resolve()



from abc import ABC, abstractmethod
from typing import Iterable

from metalpy.utils.arg_specs import ArgSpecs
from .tile_locator import TileLocator


class TileMapSource(ABC):
    def __init__(self):
        pass

    def __new__(cls, *args, **kwargs):
        arg_spec = ArgSpecs.of(cls)
        arg_spec.push_args(*args)
        arg_spec.bind_kwargs(**kwargs)
        repr_str = arg_spec.get_func_repr(cls)

        obj = super().__new__(cls)
        obj.repr = repr_str

        return obj

    @property
    def repr(self):
        return getattr(self, '__repr')

    @repr.setter
    def repr(self, val):
        setattr(self, '__repr', val)

    @abstractmethod
    def __iter__(self) -> Iterable[TileLocator]:
        """返回若干个tile资源定位器

        Returns
        -------
        iterable_of_tiles
            TileLocator的迭代器
        """
        pass

    def warp_coordinates(self, coords):
        """输入wgs84坐标系下的坐标，如果该地图源对坐标系有特殊需求，需要在此进行转换
        例如高德地图等中国地图需要转换为GCJ02坐标系

        Returns
        -------
        ret
            地图源所使用的坐标系下的坐标
        """
        return coords

    def __repr__(self):
        return self.repr



import numpy as np

from metalpy.carto.coords import Coordinates


def min_area_rect_pca(pts, debug=False):
    from sklearn.decomposition import PCA
    from scipy.spatial.transform import Rotation

    pca = PCA(n_components=pts.shape[1])
    pca.fit(pts[::2])

    rotation = np.arctan2(*pca.components_[0, ::-1])
    center = pca.mean_

    proj = Rotation.from_euler('z', -rotation).as_matrix()[:2, :2]

    # (pts - pca.mean_) @ proj.T + pca.mean_  # 将坐标点投影到新坐标系下
    size = Coordinates(pts @ proj.T).bounds.extent  # 尺寸与平移无关，只求边界时可以省略平移

    if debug:
        from matplotlib import pyplot as plt
        plt.scatter(*pts.to_numpy().T, s=0.01)
        plt.scatter(*((pts - pca.mean_) @ proj.T + pca.mean_).to_numpy().T, s=0.01)
        plt.arrow(
            *pca.mean_,
            *pca.components_[0] * 300
        )
        plt.arrow(
            *pca.mean_,
            *pca.components_[1] * 150
        )
        plt.show()

        print(np.prod(Coordinates(pts - center).bounds.extent))
        print(np.prod(Coordinates((pts - pca.mean_) @ proj.T + pca.mean_).bounds.extent))

    return center, size, rotation



from math import sqrt

import taichi as ti
from taichi import TaichiTypeError, TaichiRuntimeError

from ..solver_progress import SolverProgress


def cg(
        A: ti.linalg.LinearOperator,
        b,
        x,
        tol=1e-6,
        maxiter=5000,
        progress=False
):
    """基于线性算符的线性求解器

    基于ti.linalg.taichi_cg_solver修改而来，实现进度条，并处理精度丢失警告信息

    Parameters
    ----------
    A
        线性算符，通过LinearOperator包装
    b
        常数项
    x
        未知量，同时用于存储求解结果
    tol
        收敛容差
    maxiter
        最大迭代代数
    progress
        是否启用进度条

    Notes
    -----
    注意到正常收敛情形下， `残差的对数` 与 `迭代代数` 呈线性关系，可以作为进度条依据。

    如果残差出现波动无法正常减少，则改为基于迭代次数的进度条，以 `maxiter` 为总代数。
    """
    if b.dtype != x.dtype:
        raise TaichiTypeError(f"Dtype mismatch b.dtype({b.dtype}) != x.dtype({x.dtype}).")
    if str(b.dtype) == "f32":
        solver_dtype = ti.f32
    elif str(b.dtype) == "f64":
        solver_dtype = ti.f64
    else:
        raise TaichiTypeError(f"Not supported dtype: {b.dtype}")
    if b.shape != x.shape:
        raise TaichiRuntimeError(f"Dimension mismatch b.shape{b.shape} != x.shape{x.shape}.")

    size = b.shape
    vector_fields_builder = ti.FieldsBuilder()
    p = ti.field(dtype=solver_dtype)
    r = ti.field(dtype=solver_dtype)
    Ap = ti.field(dtype=solver_dtype)
    vector_fields_builder.dense(ti.ij, size).place(p, r, Ap)
    vector_fields_snode_tree = vector_fields_builder.finalize()

    scalar_builder = ti.FieldsBuilder()
    alpha = ti.field(dtype=solver_dtype)
    beta = ti.field(dtype=solver_dtype)
    scalar_builder.place(alpha, beta)
    scalar_snode_tree = scalar_builder.finalize()

    @ti.kernel
    def init():
        for I in ti.grouped(x):
            r[I] = b[I]
            p[I] = 0.0
            Ap[I] = 0.0

    @ti.kernel
    def reduce(p: ti.template(), q: ti.template()) -> solver_dtype:
        result = solver_dtype(0.0)
        for I in ti.grouped(p):
            result += p[I] * q[I]
        return result

    @ti.kernel
    def update_x():
        for I in ti.grouped(x):
            x[I] += alpha[None] * p[I]

    @ti.kernel
    def update_r():
        for I in ti.grouped(r):
            r[I] -= alpha[None] * Ap[I]

    @ti.kernel
    def update_p():
        for I in ti.grouped(p):
            p[I] = r[I] + beta[None] * p[I]

    def solve():
        init()
        initial_rTr = reduce(r, r)

        if progress:
            progress_bar = SolverProgress(tol, maxiter)
        else:
            progress_bar = None

        old_rTr = initial_rTr
        update_p()
        # -- Main loop --
        for i in range(maxiter):
            A._matvec(p, Ap)  # compute Ap = A x p
            pAp = reduce(p, Ap)
            alpha[None] = old_rTr / pAp
            update_x()
            update_r()
            new_rTr = reduce(r, r)
            residual = sqrt(new_rTr)

            if progress_bar is not None:
                progress_bar.sync(residual)

            if residual < tol:
                break

            beta[None] = new_rTr / old_rTr
            update_p()
            old_rTr = new_rTr

        if progress_bar is not None:
            progress_bar.close()

    solve()
    vector_fields_snode_tree.destroy()
    scalar_snode_tree.destroy()



from __future__ import annotations

import pyvista as pv

from metalpy.mepa.executor import Executor
from metalpy.utils.model import pv_ufunc_apply
from .dataset_wrapper import TDataSetWrapper
from .textures import Texture
from .texture_readers import TextureHelper
from .universal_dataset import UniversalDataSet


class TexturedDataSet(UniversalDataSet):
    def __init__(self, dataset, *args, **kwargs):
        super().__init__(dataset, *args, **kwargs)

    @classmethod
    def read(
            cls: type[TDataSetWrapper],
            model_path,
            texture=True,
            helper: TextureHelper = None
    ) -> TDataSetWrapper | pv.DataSet:
        if texture:
            helper_func = TextureHelper().bind_texture
        else:
            helper_func = None

        ret = super().read(model_path, callback=helper_func)

        return ret.view(TexturedDataSet)

    @classmethod
    def read_all(
            cls: type[TDataSetWrapper],
            path,
            progress=True,
            verbose=True,
            cache=True,
            preview: int | bool = False,
            texture=True,
            executor: Executor | None = None
    ) -> TDataSetWrapper | pv.DataSet:
        if texture:
            helper_func = TextureHelper().bind_texture
        else:
            helper_func = None

        return super().read_all(
            path,
            progress=progress,
            verbose=verbose,
            cache=cache,
            preview=preview,
            executor=executor,
            callback=helper_func,
        ).view(TexturedDataSet)

    def plot(self, texture=True, **kwargs):
        plotter = pv.Plotter()

        show_grid = kwargs.pop('show_grid', False)
        if show_grid:
            plotter.show_grid()

        show_axes = kwargs.pop('show_axes', False)
        if show_axes:
            plotter.show_axes()

        self.add_to_plotter(plotter, texture=texture, **kwargs)

        plotter.show()

    def add_to_plotter(self, plotter: pv.BasePlotter, *, texture=True, **kwargs):
        if texture:
            pv_ufunc_apply(self.dataset, lambda dataset: _add_textured_dataset_to_plotter(
                dataset,
                plotter,
                **kwargs
            ))
        else:
            plotter.add_mesh(self.dataset, **kwargs)


def _add_textured_dataset_to_plotter(dataset, plotter: pv.BasePlotter, **kwargs):
    if dataset.n_points <= 0:
        return

    textures = Texture.extract_all(dataset)
    new_kwargs = textures.build()
    new_kwargs.update(kwargs)

    actor = plotter.add_mesh(dataset, **new_kwargs)
    textures.apply(actor)



from .tile_locator import TileLocator
from ..constants import WebMercator


class SimpleTileLocator(TileLocator):
    def __init__(self, pattern: str, bottom_left_as_origin=False):
        super().__init__()
        self.pattern = pattern
        self.bottom_left_as_origin = bottom_left_as_origin

    def query(self, col, row, level) -> str:
        col, row = WebMercator.warp_tile_coord(col, row, level, self.bottom_left_as_origin)
        return self.pattern.format(
            col=col, column=col, row=row, level=level,
            x=col, y=row, z=level,  # 允许缩写
        )



from __future__ import annotations

from typing import TypeVar

import pyvista as pv

from metalpy.utils.model import DataSetLike, DataSetLikeType, pv_ufunc_map

TDataSetWrapper = TypeVar('TDataSetWrapper', bound='DataSetWrapper')


class DataSetWrapper:
    def __init__(self, dataset: DataSetLike, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if isinstance(dataset, DataSetWrapper):
            dataset = dataset.dataset

        self.dataset = dataset

    def view(self, wrapper_type: type[TDataSetWrapper]) -> TDataSetWrapper:
        return wrapper_type(self.dataset)

    def wrap(self: TDataSetWrapper, dataset: DataSetLike) -> TDataSetWrapper:
        return type(self)(dataset)

    def merge_all(self):
        # TODO: 清理 `merge` 后无用的贴图信息
        datasets = pv_ufunc_map(
            self.dataset,
            lambda dataset: dataset
        )

        return pv.merge(datasets)

    def __getattr__(self, item):
        return getattr(self.dataset, item)

    def __getitem__(self, item):
        ret = self.dataset[item]
        if isinstance(ret, DataSetLikeType):
            return self.wrap(ret)
        else:
            return ret

    def __setitem__(self, item, val):
        if isinstance(val, DataSetWrapper):
            val = val.dataset
        self.dataset[item] = val

    def __repr__(self):
        ret = self.dataset.__repr__()
        dataset_t = type(self.dataset).__name__
        this_t = type(self).__name__
        return ret.replace(
            dataset_t,
            f'{this_t}[{dataset_t}]',
        )



import warnings
from abc import ABC

from metalpy.utils.bounds import Bounds
from metalpy.utils.object_path import get_type_name
from . import Shape3D


class InfiniteShape(Shape3D, ABC):
    """用于描述一个无穷3D形状
    """

    @property
    def local_bounds(self):
        return Bounds.unbounded()

    @property
    def volume(self):
        warnings.warn(f'Trying to access `volume` of infinite shape `{get_type_name(self)}`.')
        return 0

    @property
    def area(self):
        warnings.warn(f'Trying to access `area` of infinite shape `{get_type_name(self)}`.')
        return 0

    def rotate(self, *args, **kwargs):
        warnings.warn(f'Trying to rotate infinite shape `{get_type_name(self)}`,'
                      f' which may lead to unexpected behavior related to bounds and other property.')
        return super().rotate(*args, **kwargs)

    def rotated(self, *args, **kwargs):
        warnings.warn(f'Trying to rotate infinite shape `{get_type_name(self)}`,'
                      f' which may lead to unexpected behavior related to bounds and other property.')
        return super().rotate(*args, **kwargs)

    def translate(self, *args, **kwargs):
        warnings.warn(f'Trying to translate infinite shape `{get_type_name(self)}`.')
        return super().translate(*args, **kwargs)

    def translated(self, *args, **kwargs):
        warnings.warn(f'Trying to translate infinite shape `{get_type_name(self)}`.')
        return super().translated(*args, **kwargs)



from typing import cast

import numpy as np

from metalpy.utils.numpy import FixedShapeNDArray


def _regulate_nans(val, b1, b2):
    """b1 is asserted to be shorter than b2
    """
    length = b1.n_axes * 2

    mask = np.isnan(b1)[:length]
    val[:length][mask] = b2[:length][mask]

    mask = np.isnan(b2)[:length]
    val[:length][mask] = b1[:length][mask]


def _aligned_max(b1: 'Bounds', b2: 'Bounds'):
    """b1 is asserted to be shorter than b2
    """
    length = b1.n_axes * 2

    bounds = cast(
        Bounds,
        np.array(b2, dtype=np.promote_types(b1.dtype, b2.dtype), copy=True, subok=True)
    )
    bounds[:length] = np.max([b1[:length], b2[:length]], axis=0)

    return bounds, b1.n_axes


def union(b1: 'Bounds', b2: 'Bounds'):
    b1, b2 = (b2, b1) if b1.n_axes > b2.n_axes else (b1, b2)
    bounds, n_axes = _aligned_max(b1, b2)
    bounds.origin[:n_axes] = np.min([
        b1.origin[:n_axes],
        b2.origin[:n_axes]
    ], axis=0)
    _regulate_nans(bounds, b1, b2)
    return bounds


def intersects(b1: 'Bounds', b2: 'Bounds'):
    b1, b2 = (b2, b1) if b1.n_axes > b2.n_axes else (b1, b2)
    bounds, n_axes = _aligned_max(b1, b2)
    bounds.end[:n_axes] = np.min([
        b1.end[:n_axes],
        b2.end[:n_axes]
    ], axis=0)
    _regulate_nans(bounds, b1, b2)
    return bounds


def bounded(xmin=np.nan, xmax=np.nan, ymin=np.nan, ymax=np.nan, zmin=np.nan, zmax=np.nan, *, n_axes=None):
    return Bounds.partial(
        xmin=xmin,
        xmax=xmax,
        ymin=ymin,
        ymax=ymax,
        zmin=zmin,
        zmax=zmax,
        n_axes=n_axes
    )


class Bounds(FixedShapeNDArray):
    """用于表示一个边界，接受数字和np.nan作为内容（不允许存在np.inf）

    使用np.nan用于表示在该方向上“无界”，对应到min方向上即为-inf，max方向上即为+inf
    """

    def __new__(cls, var_inp, *args, dtype=None) -> 'Bounds':
        return np.asanyarray(np.r_[var_inp, args], dtype=dtype).view(cls)

    def __array_finalize__(self, _, **__):
        pass

    @staticmethod
    def unbounded(n_axes=0) -> 'Bounds':
        """获取无界边界

        Parameters
        ----------
        n_axes
            维度

        Returns
        -------
        ret
            无界形式下的n维边界（全为np.nan）
        """
        return np.full(n_axes * 2, np.nan).view(Bounds)

    @staticmethod
    def partial(xmin=np.nan, xmax=np.nan, ymin=np.nan, ymax=np.nan, zmin=np.nan, zmax=np.nan, *, n_axes=None):
        if n_axes is None:
            n_axes = np.where(~np.isnan([0, 0, xmin, xmax, ymin, ymax, zmin, zmax]))[0][-1] // 2

        ret = Bounds.unbounded(n_axes)

        if n_axes > 0:
            ret.xmin = xmin
            ret.xmax = xmax
        if n_axes > 1:
            ret.ymin = ymin
            ret.ymax = ymax
        if n_axes > 2:
            ret.zmin = zmin
            ret.zmax = zmax

        return ret

    def as_corners(self):
        """从边界形式转换为角落点形式

        Returns
        -------
        ret
            角落点形式表示的
        """
        return self.reshape(-1, 2).T.view(Corners)

    def to_inf_format(self):
        """将无界条件从标准形式转换为正负无穷形式

        Returns
        -------
        ret
            正负无穷形式表示的边界（非标准形式）
        """
        ret = self.copy()
        infs = np.inf * np.tile([-1, 1], self.n_axes)
        mask = np.isnan(ret)
        ret[mask] = infs[mask]

        return ret

    def expand(self, *, proportion=None, increment=None, inplace=False):
        number = (int, float)
        deltas = np.tile(np.asarray((-1, 1)), self.n_axes)
        if increment is None:
            assert proportion is not None, 'Either `proportion` or `increment` must be provided.'
            if isinstance(proportion, number):
                proportion = deltas * proportion
            increment = self.extent.repeat(2) * proportion
        elif isinstance(increment, number):
            increment = increment * deltas

        if inplace:
            target = self
        else:
            target = self.copy()

        target += increment
        return target

    def override(self, by, *, inplace=False):
        """用other中的非nan值替换当前边界中的对应位置值

        Parameters
        ----------
        by
            用于覆盖self的边界值
        inplace
            指示操作是否修改self

        Returns
        -------
        ret
            覆盖后的边界对象
        """
        target = self
        if not inplace:
            target = target.copy()

        other = Bounds(by)
        length = min(target.n_axes, other.n_axes) * 2
        mask = ~np.isnan(other)[:length]
        target[:length][mask] = other[:length][mask]

        return target

    def set(self, axis, min=None, max=None):
        if np.size(min) == 2 and max is None:
            # 适配 b.set(0, [0, 10]) 用法
            min, max = min

        if min is not None:
            if np.isneginf(min):
                min = np.nan
            elif np.isposinf(min):
                raise ValueError(f'Min bound cannot be set to `+inf` in axis {axis}.')
            self[2 * axis] = min

        if max is not None:
            if np.isposinf(max):
                max = np.nan
            elif np.isneginf(max):
                raise ValueError(f'Max bound cannot be set to `-inf` in axis {axis}.')
            self[2 * axis + 1] = max

    def get(self, axis):
        return self[2 * axis: 2 * axis + 2]

    __or__ = union
    __and__ = intersects

    @property
    def extent(self):
        return self.end - self.origin

    @property
    def origin(self):
        return self[::2]

    @origin.setter
    def origin(self, val):
        self[::2] = val

    @property
    def end(self):
        return self[1::2]

    @end.setter
    def end(self, val):
        self[1::2] = val

    @property
    def xmin(self): return self[0]

    @xmin.setter
    def xmin(self, v): self[0] = v

    @property
    def xmax(self): return self[1]

    @xmax.setter
    def xmax(self, v): self[1] = v

    @property
    def xrange(self): return self[0:2]

    @xrange.setter
    def xrange(self, rng): self[0:2] = rng

    @property
    def ymin(self): return self[2]

    @ymin.setter
    def ymin(self, v): self[2] = v

    @property
    def ymax(self): return self[3]

    @ymax.setter
    def ymax(self, v): self[3] = v

    @property
    def yrange(self): return self[2:4]

    @yrange.setter
    def yrange(self, rng): self[2:4] = rng

    @property
    def zmin(self): return self[4]

    @zmin.setter
    def zmin(self, v): self[4] = v

    @property
    def zmax(self): return self[5]

    @zmax.setter
    def zmax(self, v): self[5] = v

    @property
    def zrange(self): return self[4:6]

    @zrange.setter
    def zrange(self, rng): self[4:6] = rng

    @property
    def n_axes(self): return self.shape[0] // 2

    @property
    def center(self):
        return (self.origin + self.end) / 2


class Corners(FixedShapeNDArray):
    def __new__(cls, input_arr):
        return np.asanyarray(input_arr).view(cls)

    def as_bounds(self):
        """从角落点形式转换为边界形式

        Returns
        -------
        ret
            角边界形式表示的
        """
        return self.T.ravel().view(Bounds)

    def to_inf_format(self) -> np.ndarray:
        """将无界条件从标准形式转换为正负无穷形式

        Returns
        -------
        ret
            正负无穷形式表示的角落点（非标准形式）
        """
        ret = np.array(self)
        infs = np.inf * np.asarray([-1, 1]).repeat(self.n_axes).reshape((2, -1))
        mask = np.isnan(ret)
        ret[mask] = infs[mask]

        return ret

    def set(self, axis, min=None, max=None):
        if np.size(min) == 2 and max is None:
            # 适配 c.set(0, [0, 10]) 用法
            min, max = min

        if min is not None:
            if np.isneginf(min):
                min = np.nan
            elif np.isposinf(min):
                raise ValueError(f'Min bound cannot be set to `+inf` in axis {axis}.')
            self[0, axis] = min

        if max is not None:
            if np.isposinf(max):
                max = np.nan
            elif np.isneginf(max):
                raise ValueError(f'Max bound cannot be set to `-inf` in axis {axis}.')
            self[1, axis] = max

    def get(self, axis):
        return self[:, axis]

    @property
    def extent(self):
        return self.end - self.origin

    @property
    def origin(self):
        return self[0]

    @origin.setter
    def origin(self, val):
        self[0] = val

    @property
    def end(self):
        return self[1]

    @end.setter
    def end(self, val):
        self[1] = val

    @property
    def xmin(self): return self[0, 0]

    @xmin.setter
    def xmin(self, v): self[0, 0] = v

    @property
    def xmax(self): return self[1, 0]

    @xmax.setter
    def xmax(self, v): self[1, 0] = v

    @property
    def xrange(self): return self[:, 0]

    @xrange.setter
    def xrange(self, rng): self[:, 0] = rng

    @property
    def ymin(self): return self[0, 1]

    @ymin.setter
    def ymin(self, v): self[0, 1] = v

    @property
    def ymax(self): return self[1, 1]

    @ymax.setter
    def ymax(self, v): self[1, 1] = v

    @property
    def yrange(self): return self[:, 1]

    @yrange.setter
    def yrange(self, rng): self[:, 1] = rng

    @property
    def zmin(self): return self[0, 2]

    @zmin.setter
    def zmin(self, v): self[0, 2] = v

    @property
    def zmax(self): return self[1, 2]

    @zmax.setter
    def zmax(self, v): self[1, 2] = v

    @property
    def zrange(self): return self[:, 2]

    @zrange.setter
    def zrange(self, rng): self[:, 2] = rng

    @property
    def n_axes(self): return self.origin.shape[0]

    @property
    def center(self):
        return (self.origin + self.end) / 2



from __future__ import annotations

import os
from pathlib import Path
from typing import Sequence

import tqdm
from PIL import Image, UnidentifiedImageError

from metalpy.utils.file import file_cached, PathLike, PathLikeType


@file_cached
def load_gps_info_from_exifs(paths: Sequence[PathLike] | PathLike):
    """从图像EXIF中读取位置数据信息

    Parameters
    ----------
    paths
        图像文件或所在文件夹路径，支持通过序列指定多个路径

    Returns
    -------
    exifs
        各个图像的EXIF数据
    """
    import piexif
    import pandas as pd

    if isinstance(paths, PathLikeType):
        paths = [paths]

    files_to_load = []
    for path in paths:
        if os.path.isfile(path):
            files_to_load.append(Path(path))
        else:
            for base, _, files in os.walk(path):
                base = Path(base)
                for file in files:
                    files_to_load.append(base / file)

    pts = []
    for p in tqdm.tqdm(files_to_load):
        try:
            img = Image.open(p)
        except UnidentifiedImageError:
            continue

        # TODO: other data? elevation?
        info = piexif.load(img.info['exif'])
        long = convert_rational_number(info['GPS'][piexif.GPSIFD.GPSLongitude], 60)
        long *= convert_gps_ref_code_to_sign(info['GPS'][piexif.GPSIFD.GPSLongitudeRef])
        lat = convert_rational_number(info['GPS'][piexif.GPSIFD.GPSLatitude], 60)
        lat *= convert_gps_ref_code_to_sign(info['GPS'][piexif.GPSIFD.GPSLatitudeRef])
        alt = convert_rational_number(info['GPS'][piexif.GPSIFD.GPSAltitude])
        alt *= convert_gps_ref_code_to_sign(info['GPS'][piexif.GPSIFD.GPSAltitudeRef])
        pts.append({
            'Longitude': long,
            'Latitude': lat,
            'Altitude': alt
        })

    return pd.DataFrame(pts)


def convert_rational_number(num_list, unit=10):
    if not isinstance(num_list, Sequence):
        return num_list
    if not isinstance(num_list[0], Sequence):
        num_list = (num_list,)

    p = 1
    ret = 0
    for a, b in num_list:
        ret += a / b / p
        p *= unit

    return ret


def convert_gps_ref_code_to_sign(code: str | int):
    if isinstance(code, str):
        code = code.upper()
    if code in {'E', 'N', b'E', b'N', 0}:
        return 1
    else:
        return -1



import os


class ObjSplitter:
    def __init__(self, obj_path: str):
        self.obj_path = obj_path
        self.obj_lines = []
        self.groups_index = []
        self.vertices_counts = []
        self.texture_params_counts = []
        self.normals_counts = []
        self.faces_without_offset = {}

    def __obj2lines(self):
        if not os.path.exists(self.obj_path):
            print("obj file does not exist")
            return -1
        self.obj_lines = []
        with open(self.obj_path, "r", encoding="utf-8") as f:
            self.obj_lines = f.read().splitlines()

    def __get_groups_index(self):
        # g所在位置+结尾
        self.groups_index = [index for index, line in enumerate(self.obj_lines) if "g " in line]
        self.groups_index.append(len(self.obj_lines))
        print(f"{self.groups_index=}")

    def __count_groups_coord(self):
        # 各组v,vt,vn统计
        self.vertices_counts = []
        self.texture_params_counts = []
        self.normals_counts = []
        for g_start, g_end in zip(self.groups_index, self.groups_index[1:]):
            vertices_count = sum(1 for line in self.obj_lines[g_start:g_end] if "v " in line)
            texture_params_count = sum(1 for line in self.obj_lines[g_start:g_end] if "vt " in line)
            normals_count = sum(1 for line in self.obj_lines[g_start:g_end] if "vn " in line)
            self.vertices_counts.append(vertices_count)
            self.texture_params_counts.append(texture_params_count)
            self.normals_counts.append(normals_count)
        print(f"{self.vertices_counts=}")
        print(f"{self.texture_params_counts=}")
        print(f"{self.normals_counts=}")

    def __get_faces_without_offset(self):
        self.faces_without_offset = {}
        for group_id, (g_start, g_end) in enumerate(zip(self.groups_index, self.groups_index[1:])):
            for line in self.obj_lines[g_start:g_end]:
                if "f " in line:
                    f_edges_params = line.split(" ")
                    f_edges_params_without_offset = "f"
                    for f_edge_params in f_edges_params[1:]:
                        f_edge_param = f_edge_params.split("/")
                        v = int(f_edge_param[0]) - sum(self.vertices_counts[:group_id])
                        vt = int(f_edge_param[1]) - sum(self.texture_params_counts[:group_id])
                        vn = int(f_edge_param[2]) - sum(self.normals_counts[:group_id])
                        f_edge_param_without_offset = str(v) + "/" + str(vt) + "/" + str(vn)
                        f_edges_params_without_offset = " ".join(
                            [f_edges_params_without_offset, f_edge_param_without_offset])
                    if group_id in self.faces_without_offset:
                        self.faces_without_offset[group_id].append(f_edges_params_without_offset)
                    else:
                        self.faces_without_offset[group_id] = [f_edges_params_without_offset]
        print(f"{self.faces_without_offset=}")

    def __save_rebuilt_objs(self, output_dir: str):
        # 重建分组obj
        for group_id, (g_start, g_end) in enumerate(zip(self.groups_index, self.groups_index[1:])):
            f_first_index = g_start + next(
                index for index, line in enumerate(self.obj_lines[g_start:g_end]) if "f " in line)
            origin = "\n".join(self.obj_lines[g_start:f_first_index])
            faces = "\n".join(self.faces_without_offset[group_id])
            os.makedirs(output_dir, exist_ok=True)
            with open(os.path.join(output_dir, f"{group_id}.obj"), "w", encoding="utf-8") as output:
                output.write("\n".join([origin, faces]))

    def split_by_group(self, output_dir: str):
        err = self.__obj2lines()
        if err:
            return
        self.__get_groups_index()
        self.__count_groups_coord()
        self.__get_faces_without_offset()
        self.__save_rebuilt_objs(output_dir)


if __name__ == "__main__":
    obj_splitter = ObjSplitter("origin.obj")
    obj_splitter.split_by_group("./output")



from __future__ import annotations

from abc import ABC, abstractmethod
from functools import reduce
from typing import Iterable, TYPE_CHECKING

import numpy as np
import tqdm

from metalpy.utils.dhash import dhash
from metalpy.utils.bounds import Bounds, union
from ..transform import CompositeTransform, Transform, Translation, Rotation
from ..utils.mesh import is_inside_bounds

if TYPE_CHECKING:
    from metalpy.scab.modelling import Scene
    from metalpy.scab.modelling.modelled_mesh import ModelledMesh
    import pyvista as pv


class Shape3D(ABC):
    def __init__(self):
        self.transforms = CompositeTransform()

    @property
    def prune_mesh(self):
        """指示在子类do_place前自动根据Bounds框裁剪边界之外的网格
        """
        return True

    def place(self, mesh_cell_centers, progress=False):
        """计算模型体所占用的网格

        Parameters
        ----------
        mesh_cell_centers
            3维网格中心点坐标
        progress
            进度条实例或None，当子类的n_task属性不为1时，需自行更新进度条

        Returns
        -------
        ret
            布尔数组或浮点数组，浮点数组原则上来说取值范围应为[0, 1]，
            指示对应网格位置是否有效或有效的程度， 0 代表非活动网格

        Notes
        -----
        内部同时负责对坐标点进行变换并进行裁剪，保证`do_place`内部只需要按局部坐标系计算
        """
        progress = self._check_progress(progress)

        mesh_cell_centers = self.before_place(mesh_cell_centers, progress)

        indices = None
        if self.prune_mesh:
            indices = is_inside_bounds(mesh_cell_centers, self.local_bounds)
            mesh_cell_centers = mesh_cell_centers[indices]

        ind = self.do_place(mesh_cell_centers, progress)

        if indices is not None:
            indices[indices] = ind
        else:
            indices = ind

        return indices

    def compute_implicit_distance(self, mesh_cell_centers, progress=False):
        """计算模型体到空间中任一点的有向距离，且小于0代表在模型内

        Parameters
        ----------
        mesh_cell_centers
            3维网格中心点坐标
        progress
            进度条实例或None，当子类的n_task属性不为1时，需自行更新进度条

        Returns
        -------
        ret
            浮点数组，指示空间中任意一点到模型体表面的有向距离，且小于0代表在模型内

        Notes
        -----
        内部同时负责对坐标点进行变换，保证`do_compute_implicit_distance`内部只需要按局部坐标系计算
        """
        progress = self._check_progress(progress)

        mesh_cell_centers = self.before_place(mesh_cell_centers, progress)
        return self.do_compute_implicit_distance(mesh_cell_centers, progress)

    def before_place(self, mesh_cell_centers, progress):
        return self.transforms.inverse_transform(mesh_cell_centers)

    def do_place(self, mesh_cell_centers, progress):
        """计算模型体所占用的网格的实现函数，所有Shape3D的子类应当重写该函数

        默认实现依赖to_local_polydata()，基于PyVista的select_enclosed_points，存在性能问题

        Parameters
        ----------
        mesh_cell_centers
            3维网格中心点坐标
        progress
            进度条实例或None，当子类的n_task属性不为1时，需自行更新进度条

        Returns
        -------
        ret
            布尔数组或浮点数组，浮点数组原则上来说取值范围应为[0, 1]，
            指示对应网格位置是否有效或有效的程度，0代表非活动网格

        Notes
        -----
        在Scene.build_mesh_worker中假定了0代表非活动网格，不参与模型间重叠部分的计算，请避免将0作为一个有意义的值输出

        外部`place`通过`before_place`对坐标点进行了逆变换，只需要按局部坐标系计算即可
        """
        import pyvista as pv

        mesh = pv.PolyData(mesh_cell_centers).cast_to_unstructured_grid()
        surface = self.to_local_polydata().extract_surface()

        selection = mesh.select_enclosed_points(surface, tolerance=0.0, check_surface=True)

        if progress and self.progress_manually:
            progress.update(self.n_tasks)

        return selection['SelectedPoints'].view(bool, np.ndarray)

    def do_compute_implicit_distance(self, mesh_cell_centers, progress):
        """计算模型体到空间中任一点的有向距离的具体实现

        默认实现依赖to_local_polydata()，基于PyVista的compute_implicit_distance，
        需要进一步优化可以重载此函数

        Parameters
        ----------
        mesh_cell_centers
            3维网格中心点坐标
        progress
            进度条实例或None，当子类的n_task属性不为1时，需自行更新进度条

        Returns
        -------
        ret
            浮点数组，指示空间中任意一点到模型体表面的有向距离，且小于0代表在模型内

        Notes
        -----
        外部`compute_implicit_distance`通过`before_place`对坐标点进行了逆变换，只需要按局部坐标系计算即可
        """
        import pyvista as pv

        poly = pv.PolyData(mesh_cell_centers)
        surface = self.to_local_polydata().extract_surface()

        poly = poly.compute_implicit_distance(surface, inplace=True)

        if progress and self.progress_manually:
            progress.update(self.n_tasks)

        return poly['implicit_distance'].view(np.ndarray)

    def __hash__(self):
        return dhash(self).digest()

    @abstractmethod
    def __dhash__(self):
        return dhash(self.transforms)

    def clone(self, deep=True):
        ret = self.do_clone(deep=deep)
        ret.transforms = self.transforms.clone()
        return ret

    @abstractmethod
    def do_clone(self, deep=True):
        raise NotImplementedError()

    def to_polydata(self) -> 'pv.PolyData':
        """将 Shape 转换为 PolyData 表示

        Returns
        -------
        ret
            PolyData 表示的 Shape

        Notes
        -----
            to_polydata 会在局部坐标系下的 PolyData 上应用变换，映射得到世界坐标系下的 PolyData

            因此继承类应当实现 to_local_polydata
        """
        import pyvista as pv
        from metalpy.utils.model import pv_ufunc_apply

        ret: pv.PolyData | pv.MultiBlock = self.to_local_polydata()

        if ret is not None:
            def transform(model):
                pts = self.transforms.transform(model.points)
                model.points = pts
            pv_ufunc_apply(ret, transform, inplace=True)

        return ret

    def to_local_polydata(self) -> 'pv.PolyData':
        """将 Shape 转换为本地坐标系下的 PolyData 表示

        Returns
        -------
        ret
            本地坐标系下 PolyData 表示的 Shape

        Notes
        -----
            继承类应当实现 to_local_polydata ，然后由 to_polydata 映射到世界坐标系中
        """
        raise NotImplementedError()

    def to_scene(self, model=True) -> 'Scene':
        """直接基于当前Shape3D实例构建场景

        Parameters
        ----------
        model
            模型值，默认为True，代表有效

        Returns
        -------
        scene
            仅包含当前Shape3D实例的场景Scene

        See Also
        --------
        :class:`metalpy.scab.modelling.Scene` : 用于建模的场景对象
        """
        from metalpy.scab.modelling import Scene
        return Scene.of(self, models=(model,), skip_checking=True)

    def build(self, model=True, **kwargs) -> 'ModelledMesh':
        """直接基于当前Shape3D实例构建场景的网格和模型

        Parameters
        ----------
        model
            模型值，默认为True，代表有效
        kwargs
            其它适用于Scene.build的参数

        Returns
        -------
        model_mesh
            模型网格

        See Also
        --------
            Scene.build : 构建场景的网格和模型
        """
        return self.to_scene(model=model).build(**kwargs)

    def plot(self, *args, **kwargs):
        """直接生成PyVista对象并调用plot在窗口中绘制

        Parameters
        ----------
        kwargs
            其它适用于pv.PolyData.plot的参数

        See Also
        --------
            Shape3D.to_polydata : 生成PyVista对象
        """
        self.to_polydata().plot(*args, **kwargs)

    @property
    def center(self) -> np.ndarray:
        """获取Shape在世界坐标系下的中心点坐标

        Returns
        -------
        ret : array(3)
            Shape在世界坐标系下的中心点坐标[x, y, z]
        """
        return self.transforms.transform(self.local_center)

    @property
    def bounds(self) -> Bounds:
        """获取Shape在世界坐标系下的长方体包围盒

        Returns
        -------
        ret : array(6)
            Shape在世界坐标系下的长方体包围盒[x0, x1, y0, y1, z0, z1]
        """
        bounds = self.oriented_bounds

        ret = Bounds.unbounded(3)
        for i in range(ret.n_axes):
            val = bounds[:, i]
            val = val[~np.isnan(val)]

            if len(bounds) == 0:
                # 无上下界
                continue

            ret.set(i, min=val.min(), max=val.max())

        return Bounds(ret)

    @property
    def oriented_bounds(self) -> np.ndarray:
        """获取Shape在世界坐标系下的八点包围盒

        Returns
        -------
        ret : array(8, 3)
            Shape在世界坐标系下的八点包围盒[[x0, y0, z0], ...[x7, y7, z7]]

        Notes
        -----
        八点包围盒表示法下，无界会使用对应的正负无穷（inf）表示（Bounds约定nan为无界）

        各种空间变换由于精度问题，对无界 / 正负无穷（inf）边界的变换结果可能无法预测
        """
        bounds = self.local_oriented_bounds
        bounds = self.transforms.transform(bounds)

        return bounds

    @property
    def local_center(self) -> np.ndarray:
        """获取Shape在局部坐标系下的中心点

        Returns
        -------
        ret : array(3)
            Shape在局部坐标系下的中心点坐标[x, y, z]
        """
        bounds = self.local_oriented_bounds
        return bounds.mean(axis=0)

    @property
    def local_bounds(self) -> Bounds:
        """获取Shape在局部坐标系下的长方体包围盒

        默认实现基于to_local_polydata，存在性能问题，
        子类应当实现local_oriented_bounds和local_bounds中的一个

        Returns
        -------
        ret : array(6)
            Shape在局部坐标系下的长方体包围盒[x0, x1, y0, y1, z0, z1]
        """
        if type(self).local_oriented_bounds is Shape3D.local_oriented_bounds:
            return Bounds(self.to_local_polydata().bounds)
        else:
            bounds = self.local_oriented_bounds
            ret = np.zeros(6)
            ret[::2] = np.min(bounds, axis=1)
            ret[1::2] = np.max(bounds, axis=1)

            return Bounds(ret)

    @property
    def local_oriented_bounds(self) -> np.ndarray:
        """获取Shape在局部坐标系下的八点包围盒

        默认实现基于to_local_polydata，存在性能问题，
        子类应当实现local_oriented_bounds和local_bounds中的一个

        Returns
        -------
        ret : array(8, 3)
            Shape在局部坐标系下的八点包围盒[[x0, y0, z0], ...[x7, y7, z7]]

        Notes
        -----
        八点包围盒表示法下，无界会使用对应的正负无穷（inf）表示（Bounds约定nan为无界）
        """
        infs = [np.inf, -np.inf]

        local_bounds = self.local_bounds.to_inf_format()
        n_axes = local_bounds.n_axes

        xrng = local_bounds.xrange if n_axes > 0 else infs
        yrng = local_bounds.yrange if n_axes > 1 else infs
        zrng = local_bounds.zrange if n_axes > 2 else infs

        x, y, z = np.meshgrid(xrng, yrng, zrng, indexing='ij')
        return np.c_[x.ravel(), y.ravel(), z.ravel()]

    def apply(self, trans: Transform, inplace=False):
        """逻辑上对空间体位置进行变换，目前通过对网格点进行逆变换实现

        Parameters
        ----------
        trans
            待应用的变换
        inplace
            指示操作是否应用在当前实例上

        Returns
        -------
        ret
            当inplace为True，返回当前实例，否则返回一个变换后的新对象
        """
        if inplace:
            ret = self
        else:
            ret = self.clone()

        ret.transforms.add(trans)
        return ret

    def translate(self, x, y, z, inplace=False):
        """对Shape进行平移

        Parameters
        ----------
        x, y, z
            三个方向的位移量
        inplace
            指示操作是否应用在当前实例上

        Returns
        -------
        ret
            当inplace为True，返回当前实例，否则返回一个平移后的新对象
        """
        return self.apply(Translation(x, y, z), inplace=inplace)

    def rotate(self, y, a, b, degrees=False, seq='xyz', inplace=False):
        """对Shape进行旋转，方向遵循右手准则

        Parameters
        ----------
        seq
            旋转顺序
        y, a, b
            对应于seq的旋转量
        degrees
            指示y，a，b是否为角度制
        inplace
            指示操作是否应用在当前实例上

        Returns
        -------
        ret
            当inplace为True，返回当前实例，否则返回一个旋转后的新对象
        """
        return self.apply(Rotation(y, a, b, degrees=degrees, seq=seq), inplace=inplace)

    def translated(self, x, y, z, inplace=True):
        """对Shape进行平移

        Parameters
        ----------
        x, y, z
            三个方向的位移量
        inplace
            指示操作是否应用在当前实例上

        Returns
        -------
        ret
            默认应用与当前对象，并返回当前对象
        """
        return self.translate(x, y, z, inplace=inplace)

    def rotated(self, y, a, b, degrees=False, seq='xyz', inplace=True):
        """对Shape进行旋转，方向遵循右手准则

        Parameters
        ----------
        seq
            旋转顺序
        y, a, b
            对应于seq的旋转量
        degrees
            指示y，a，b是否为角度制
        inplace
            指示操作是否应用在当前实例上

        Returns
        -------
        ret
            默认应用与当前对象，并返回当前对象
        """
        return self.rotate(y, a, b, degrees=degrees, seq=seq, inplace=inplace)

    @property
    def volume(self):
        """获取该几何体的体积
        """
        raise NotImplementedError()

    @property
    def area(self):
        """获取该几何体的外表面积
        """
        raise NotImplementedError()

    @property
    def n_tasks(self):
        """指示子类的细分任务数。

        当返回值不为1时，子类可选择自行更新进度条，或重载progress_manually来指定是否自行更新进度条。

        Notes
        -----
        注意，如果n_tasks为动态值且可能为1，
        则可能需要手动重载progress_manually来协助Scene明确是否需要在Shape3D外进行更新
        """
        return 1

    @property
    def progress_manually(self):
        """指示子类是否自行更新进度条
        """
        return self.n_tasks != 1

    def _check_progress(self, progress):
        if progress is True and self.progress_manually:
            return tqdm.tqdm(total=self.n_tasks)
        elif progress:
            return progress
        else:
            return None


def bounding_box_of(shapes: Iterable[Shape3D]):
    return reduce(union, (shape.bounds for shape in shapes))



from __future__ import annotations

import warnings
from typing import Iterable

import numpy as np
import pandas as pd
import tqdm
from scipy import stats

from metalpy.utils.file import file_cached
from metalpy.utils.math import weighted_percentile


def check_slope(x, y):
    """检查x和y的方差，变化更大的方向作为x，将计算的斜率定义域限定在[-pi/2, pi/2]

    Parameters
    ----------
    x, y
        点位置的x、y坐标

    Returns
    -------
    x, y
        如果x方向方差大于y，则返回x，y，否则搜索

    Notes
    -----
    该函数假定了点位置中绝大多数时间航向相同或相反
    """
    mid = len(x) // 2
    if np.std(x[mid - 50:mid + 50]) < np.std(y[mid - 50:mid + 50]):
        return y, x
    else:
        return x, y


def apply_delta_to_segments(segments, delta):
    return [(a + delta, b + delta) for a, b in segments]


def regulate_delta_directions(directions, inplace=False):
    if np.ndim(directions) == 0:
        if inplace:
            warnings.warn('Can not operate inplace-ly on a single element'
                          ' (with `inplace = True`).')
        if directions > 90:
            directions -= 180
        if directions < -90:
            directions += 180
        return directions

    if not inplace:
        directions = np.copy(directions)

    directions[directions > 90] -= 180
    directions[directions < -90] += 180  # 保证偏转角连续

    return directions


def diff_directions(directions):
    diff_direction = np.diff(directions)
    regulate_delta_directions(diff_direction)

    return diff_direction


@file_cached
def analyze_directions(x, y,
                       slices: Iterable[list | tuple | slice] | None = None,
                       robust=False,
                       progress=False):
    if slices is None:
        slices = [slice(None)]

    if progress:
        slices = tqdm.tqdm(slices)

    directions = []
    for s in slices:
        if isinstance(s, (tuple, list)):
            s = slice(*s)

        if np.unique(x[s], axis=0).shape[0] == 1:
            unique_ys = np.unique(y[s], axis=0)
            n_ys = unique_ys.shape[0]
            if n_ys == 1 and len(directions) > 0:
                # 悬停
                directions.append(directions[-1])
            elif n_ys > 2:
                # 需要至少三个不同y才能分辨悬停和垂直航线
                directions.append(np.pi / 2)
            else:
                directions.append(np.nan)
        else:
            if not robust:
                slope, intercept, r_value, p_value, std_err = stats.linregress(x[s], y[s])
            else:
                slope, intercept = stats.siegelslopes(y[s], x[s])
            directions.append(np.arctan(slope))

    return directions


def prune_by_std(directions) -> tuple:
    """筛选掉航段两端航向偏差较大的异常值（通常为转向带来的误差）

    Parameters
    ----------
    directions
        飞机航向角

    Returns
    -------
    sub_range
        返回筛选掉两端异常值后的区间
    """
    n_pts = len(directions)
    std = directions.std()
    center = np.percentile(directions, 50)

    deltas = abs(directions - center)
    is_normal = deltas <= 2 * std

    dl = np.argmax(is_normal)
    dr = np.argmax(is_normal[::-1])

    return dl, n_pts - dr


def segment_by_std(directions, stable_window_size) -> list[tuple]:
    """假设该航段由两个航段和一个连续的转弯组成，则寻找某个划分点将其分成两部分，且两部分为独立的两个航段，因此应方差相近

    Parameters
    ----------
    directions
        飞机航向角
    stable_window_size
        用于确定为航线的窗口测点数。

    Returns
    -------
    sub_segments
        如果可以切分，则切分为两段，并通过stable_window_size筛选掉过短的段。
        否则不分段直接返回 [(0, n_pts)]。
    """
    n_pts = len(directions)
    left, right = 0, n_pts
    lstd, rstd, mid = 100, -100, -1

    while right - left > 1:
        mid = (left + right) // 2
        lstd, rstd = directions[:mid].std(), directions[mid:].std()
        if lstd > rstd:
            right = mid
        else:
            left = mid

    if abs(lstd - rstd) < 0.1:
        ret = (0, mid), (mid, n_pts)
    else:
        ret = (0, n_pts),

    ret = [r for r in ret if r[1] - r[0] > stable_window_size]

    return ret


def segment_by_diff_directions(directions, stable_window_size, stable_direction_tol):
    diff_direction = np.r_[diff_directions(directions), 1e9]  # 保证最后一个航段也会统计进来

    segments = []
    first = None
    for i in range(len(diff_direction)):
        ddi = diff_direction[i]
        if abs(ddi) < stable_direction_tol:
            if first is None:
                first = i
        else:
            if first is not None:
                if i - first > stable_window_size:
                    segments.append((first, i))
                first = None

    new_segments = []
    for a, b in segments:
        ds = directions[a:b]
        if ds.std() > stable_direction_tol:
            new_segs = segment_by_std(ds, stable_window_size)
            new_segments.extend(apply_delta_to_segments(new_segs, a))
        else:
            new_segments.append((a, b))
    segments = new_segments

    new_segments = []
    for a, b in segments:
        new_segs = prune_by_std(directions[a:b])
        new_segments.extend(apply_delta_to_segments((new_segs,), a))
    segments = new_segments

    return segments


def split_lines(
        x, y,
        window_size=None,
        stable_window_size=None,
        window_length=10,
        stable_window_length=50,
        stable_direction_tol: float = 3,
):
    """通过滑动窗口求取区间航向，并依据航向的变化幅度分割为若干测线

    Parameters
    ----------
    x, y
        航线水平路径
    window_size, window_length
        用于分析航向的窗口测点数。应当大于飞机产生明显位移的窗口大小。
        过小时可能会导致测线划分对飞机飞行扰动敏感产生错误划分，过大时可能导致转弯等非平稳飞行测点也被容纳进测线。
    stable_window_size
        用于确定为航线的窗口测点数。应当大于转弯等非平稳飞行过程的窗口大小。
        过小时可能导致转弯等非平稳测点被统计为测线，过大时可能导致某些较短测线被意外移除。
    window_length, stable_window_length
        对应的窗口长度。若不指定对应size而指定length，则自动统计测点间距然后通过length求取窗口测点数。
    stable_direction_tol
        航向角偏差容差（单位 `度` ）

    Returns
    -------
    segments
        初步拆分后的若干航段
    """
    n_pts = len(x)

    if window_size is None or stable_window_size is None:
        skip = 2
        dx, dy = np.diff(x[::skip]), np.diff(y[::skip])
        d = np.sqrt(np.percentile(dx ** 2 + dy ** 2, 50)) / skip

        min_group_points = 5
        if window_size is None:
            window_size = max(int(np.ceil(window_length / d)), min_group_points)
        if stable_window_size is None:
            stable_window_size = max(int(np.ceil(stable_window_length / d)), min_group_points)

    x, y = check_slope(x, y)

    windows = [(idx, idx + window_size) for idx in range(n_pts - window_size)]
    directions = analyze_directions(x, y, windows, progress=True)
    directions = np.rad2deg(directions)

    segments = segment_by_diff_directions(directions, stable_window_size, stable_direction_tol)
    segments = [(first, last + window_size + 1) for first, last in segments]

    return segments


def analyze_lines(x, y, segments):
    details = analyze_lines_details(x, y, segments)
    return details.slope, details.intercept


def analyze_lines_details(x, y, segments):
    x, y = check_slope(x, y)

    ret = []
    for first, last in segments:
        sx = x[first:last]
        sy = y[first:last]
        result = stats.linregress(sx, sy)
        ret.append(result)

    return pd.DataFrame(ret)


def remove_aux_flights(x, y, segments, stable_threshold: float = 6):
    directions = analyze_directions(x, y, segments, robust=True)
    analyze_directions(x, y, [segments[-1]], robust=True)
    weights = [s[1] - s[0] for s in segments]

    aux_removed_idx = remove_aux_flights_by_directions(directions, weights, stable_threshold=stable_threshold)

    aux_removed = []
    for idx in aux_removed_idx:
        aux_removed.append(segments[idx])

    return aux_removed


def remove_aux_flights_by_directions(directions, weights, stable_threshold: float = 6):
    directions = np.rad2deg(directions)
    d_dir = directions - directions[0]
    d_dir = regulate_delta_directions(d_dir, inplace=True)

    selected_direction = weighted_percentile(d_dir, weights, 50) + directions[0]
    selected_direction = regulate_delta_directions(selected_direction)

    angle_offsets = abs(regulate_delta_directions(directions - selected_direction))
    condition = angle_offsets < stable_threshold

    aux_removed = []

    for idx in np.where(condition)[0]:
        aux_removed.append(idx)

    return aux_removed



from functools import lru_cache

import numpy as np
from pyproj import CRS, Transformer

from metalpy.utils.bounds import Bounds


class WebMercator:
    """适用于四叉树组织的WebMercator坐标系相关计算

    TODO：实现二分-四叉树组织相关计算
    """
    DirectionAll = 'all'
    WGS84 = CRS.from_string('WGS 84')
    PseudoMercator = CRS.from_string('WGS 84 / Pseudo-Mercator')

    Perimeter = 2 * np.pi * PseudoMercator.ellipsoid.semi_major_metre
    WGS84_PseudoMercator = Transformer.from_crs(WGS84, PseudoMercator, always_xy=True)
    PseudoMercator_WGS84 = Transformer.from_crs(PseudoMercator, WGS84, always_xy=True)

    @staticmethod
    def wgs84_to_pseudo_mercator(long, lat):
        """从WGS84转换为Pseudo-Mercator坐标系
        """
        x, y = WebMercator.WGS84_PseudoMercator.transform(long, lat)

        return x, y

    @staticmethod
    def pseudo_mercator_to_wgs84(x, y):
        """从Pseudo-Mercator转换为WGS84坐标系
        """
        long, lat = WebMercator.PseudoMercator_WGS84.transform(x, y)

        return long, lat

    @staticmethod
    @lru_cache
    def tiles_in_axis(direction, level):
        """计算指定方向上在level等级下的tile数

        Parameters
        ----------
        direction
            方向，0代表x东西，1代表y南北
        level
            tile等级

        Returns
        -------
        ret
            在level等级下给定方向上最大tile数
        """
        return 2 ** level

    @staticmethod
    def tile_width(direction, level):
        return WebMercator.Perimeter / WebMercator.tiles_in_axis(direction, level)

    @staticmethod
    def pseudo_mercator_bounds_to_tile(bounds, level):
        """将Pseudo-Mercator坐标边界转换为Tile序号边界。

        Parameters
        ----------
        bounds
            Pseudo-Mercator坐标边界
        level
            缩放等级

        Returns
        -------
        tile_bounds
            Tile序号边界

        Notes
        -----
        由于原点方向不一致，转换后y方向上下边界会交换，例如TileYMax对应MercatorYMin。
        """
        tile_width = WebMercator.tile_width(WebMercator.DirectionAll, level)
        y_tiles = WebMercator.tiles_in_axis(WebMercator.DirectionAll, level)

        ret = np.floor(Bounds(bounds + WebMercator.Perimeter / 2) / tile_width).astype(np.int32)
        ret.ymin, ret.ymax = y_tiles - ret.ymax - 1, y_tiles - ret.ymin - 1

        return ret

    @staticmethod
    def tile_bounds_to_pseudo_mercator(bounds, level):
        """将Tile序号边界转换为Pseudo-Mercator坐标边界

        Parameters
        ----------
        bounds
            Tile序号边界
        level
            缩放等级

        Returns
        -------
        pseudo_mercator_bounds
            Pseudo-Mercator坐标边界

        Notes
        -----
        由于tile坐标指向的为一个瓦片而非点，因此边界转换时会取最小位置瓦片的下边界和最大位置瓦片的上边界

        由于原点方向不一致，转换后y方向上下边界会交换，例如MercatorYMin对应TileYMax。
        """
        tile_width = WebMercator.tile_width(WebMercator.DirectionAll, level)
        y_tiles = WebMercator.tiles_in_axis(WebMercator.DirectionAll, level)

        bounds = Bounds(bounds).copy()
        bounds.ymin, bounds.ymax = y_tiles - bounds.ymax - 1, y_tiles - bounds.ymin - 1

        return (bounds + [0, 1, 0, 1]) * tile_width - WebMercator.Perimeter / 2

    @staticmethod
    def to_bounding_tiles_mercator_bounds(mercator_bounds, level):
        """将Pseudo-Mercator边界坐标扩大为所涉及的Tile的Pseudo-Mercator边界
        
        Parameters
        ----------
        mercator_bounds
            需要的地图边界
        level
            缩放等级

        Returns
        -------
        tiles_mercator_bounds
            所涉及Tile的Pseudo-Mercator边界
        """
        tile_bounds = WebMercator.pseudo_mercator_bounds_to_tile(mercator_bounds, level)
        tiles_mercator_bounds = WebMercator.tile_bounds_to_pseudo_mercator(tile_bounds, level)

        return tiles_mercator_bounds

    @staticmethod
    def iter_tiles(mercator_bounds, level):
        mercator_bounds = np.array(mercator_bounds)
        tile_bounds = WebMercator.pseudo_mercator_bounds_to_tile(mercator_bounds, level)

        for x in range(tile_bounds[0], tile_bounds[1] + 1):
            for y in range(tile_bounds[2], tile_bounds[3] + 1):
                yield x, y

    @staticmethod
    def warp_tile_coord(x, y, z, bottom_left_as_origin=False):
        """WMTS规定北纬85.05°为零点，该函数用于根据给定参数，转换为对应的南纬85.05°为零点或不转换

        Parameters
        ----------
        x,y
            左下原点标准下x、y方向tile坐标
        z
            tile等级
        bottom_left_as_origin
            是否采用左下角作为tile原点

        Returns
        -------
        tile_coord_in_top_left_as_origin
            给定约定下的tile坐标
        """
        if bottom_left_as_origin:
            y = WebMercator.tiles_in_axis(1, z) - y - 1
        return x, y



from . import Coefficient


class Alpha(Coefficient):
    @property
    def name(self):
        return 'opacity'

    def build(self):
        if self.texture == 1:
            return {}
        else:
            return super().build()



import numpy as np

from metalpy.scab.modelling.shapes import Shape3D


def assert_placing_inside(shape: Shape3D, cell_size=1):
    """验证 `shape.place` 计算的所有有效网格均在模型体内。

    通过PyVista的体素化方法或隐式距离方法进行验证。
    """
    model_mesh = shape.build(cell_size=cell_size)

    try:
        active = Shape3D.do_place(shape, model_mesh.active_cell_centers, None)
    except (RuntimeError, AttributeError):
        # shape为MultiBlock或者其它异常情况导致默认的 `do_place` 无效
        active = shape.compute_implicit_distance(model_mesh.active_cell_centers) <= 0

    assert np.all(active)



from __future__ import annotations

from math import log10, inf

import tqdm

from metalpy.utils.numeric import limit_significand


class SolverProgress:
    def __init__(self, tol, maxiter, *, divergence_tol=27, mapper=log10, unit='logRES'):
        """通过残差来计算进度

        默认采用残差对数，不难注意到一些优化算法下，残差的对数呈线性下降趋势

        Parameters
        ----------
        tol
            目标残差
        maxiter
            最大运行代数
        divergence_tol
            发散检测阈值，当残差大于初始残差次数超过 `divergence_tol` 时，改为按迭代代数统计进度
        mapper
            残差转换函数，默认为限制有效位数下的对数函数
        unit
            残差单位，如果修改了 `mapper` 应该也要改一下单位
        """
        self.mapper = mapper
        self.unit = unit

        self.postfixes: dict[str, float | str] = {
            'iter': 0,
            'residual': inf,
            'target': tol
        }
        self._n_iter = 0
        self.log_res = 0
        self.progress_bar = None
        self.progress_by_res_counter = divergence_tol

        self.end = self._update_residual(tol)
        self.origin = 0
        self.maxiter = maxiter

    def sync(self, residual):
        self.n_iter += 1
        log_res = self._update_residual(residual)

        if self.progress_bar is None:
            # 先等待残差下降才启动进度条
            self.origin = origin = log_res
            self.progress_bar = tqdm.tqdm(
                total=origin - self.end,
                unit=self.unit,
                postfix=self.postfixes
            )
        elif self.progress_by_res_counter >= 0:
            total = self.origin - log_res
            if total < 0:
                if self.progress_by_res_counter > 0:
                    total = 0
                self.progress_by_res_counter -= 1
            if total >= 0:
                if total > self.total:
                    total = self.total
                delta = total - self.progress_bar.n
                self.progress_bar.set_postfix(self.postfixes, refresh=False)
                self.progress_bar.update(delta)
            else:
                # 就要发散就要发散，给您改为基于迭代代数的进度条
                self.postfixes.pop('iter')  # 不需要再在后缀中展示当前代数
                self.progress_bar.unit = 'it'
                self.progress_bar.total = self.maxiter
                self.progress_bar.last_print_n = self.n_iter
                self.progress_bar.n = self.n_iter + 1
                self.progress_bar.set_postfix(self.postfixes, refresh=False)
                self.progress_bar.refresh()
        else:
            self.progress_bar.set_postfix(residual=self.res_str, refresh=False)
            self.progress_bar.update(1)

    def close(self):
        if self.progress_bar is not None:
            self.progress_bar.close()

    @property
    def n_iter(self):
        return self._n_iter

    @n_iter.setter
    def n_iter(self, v):
        self._n_iter = v
        if 'iter' in self.postfixes:
            self.postfixes['iter'] = v

    @property
    def res_str(self):
        return self.postfixes['residual']

    @res_str.setter
    def res_str(self, v):
        self.postfixes['residual'] = v

    @property
    def total(self):
        return self.progress_bar.total

    def _update_residual(self, residual):
        log_res = limit_significand(self.mapper(residual), bits=6)
        res_str = f'{residual:.2e}'

        self.res_str = res_str
        self.log_res = log_res

        return log_res



from __future__ import annotations

from typing import Iterable, overload

from SimPEG import maps
from SimPEG.potential_fields import magnetics

from metalpy.scab.utils.misc import define_inducing_field
from .. import BasePFSimulationBuilder
from ... import SimulationBuilder


class Simulation3DIntegralBuilder(BasePFSimulationBuilder):
    def __init__(self, sim_cls: magnetics.Simulation3DIntegral):
        super().__init__(sim_cls)
        self._receiver_list = []
        self._source_field = None
        self._model_type = None

        self.source_field(50000, 90, 0)
        self.scalar_model()

    def build(self) -> magnetics.Simulation3DIntegral:
        return super().build()

    def receivers(self, receiver_points, components: str | Iterable[str] = 'tmi'):
        self._receiver_list.append(
            magnetics.receivers.Point(
                receiver_points, components=components
            ))

    def source_field(self, strength, inc, dec):
        self._source_field = define_inducing_field(strength, inc, dec)

    @SimulationBuilder._assembles('survey')
    def _survey(self):
        field = magnetics.sources.UniformBackgroundField(
            receiver_list=self._receiver_list,
            amplitude=self._source_field.strength,
            inclination=self._source_field.inclination,
            declination=self._source_field.declination,
        )
        survey = magnetics.survey.Survey(field)

        return survey

    @overload
    def model_type(self, model_type: str): ...

    @overload
    def model_type(self, *, scalar: bool): ...

    @overload
    def model_type(self, *, vector: bool): ...

    @SimulationBuilder._supplies('model_type')
    def model_type(self, type=None, *, scalar=None, vector=None):
        """设置模型类型，对应Simulation3DIntegral的model_type参数。

        `scalar`代表标量模型，输入为每个网格的磁化率。

        `vector`代表向量模型，输入为每个网格的等效磁化率向量，
        为(n_cells, 3)形式模型的按Fortran序展开（flatten(order='F')）
        或使用discretize.utils的mkvc。
        """
        assert (type is None) + (scalar is None) + (vector is None) == 2, \
            'Only one of `model_type`, `scalar` or `vector` can be provided for `model_type()`.'
        if type is not None:
            pass
        elif scalar is True:
            type = 'scalar'
        elif vector is True:
            type = 'vector'

        self._model_type = type
        return type

    def scalar_model(self):
        return self.model_type(scalar=True)

    def vector_model(self):
        return self.model_type(vector=True)

    @SimulationBuilder._supplies('chiMap')
    def chi_map(self, chi_map: maps.IdentityMap):
        return chi_map

    @SimulationBuilder._assembles('chiMap')
    def _chi_map(self):
        # TODO: 考察是否需要默认提供 maps.ChiMap （从磁化率转换为磁导率，数值上约相差1）
        #   需要注意，向量模型不可直接使用 maps.ChiMap
        n_params = self._model_mesh.n_active_cells
        if self._model_type == 'vector':
            n_params *= 3
        return maps.IdentityMap(nP=n_params)



import itertools
from typing import Any

from metalpy.utils.arg_specs import ArgSpecs
from metalpy.utils.type import undefined


class LazyClassFactory:
    def __init__(self, cls, *args, _argspecs: ArgSpecs = None, **kwargs):
        """用于实现类的延迟构造，
        通过construct可以使用对应的类进行构造，new_kwargs指定用于替换或追加的参数，
        通过clone可以复制一个新的延迟构造器

        Parameters
        ----------
        cls
            待构造目标类
        _argspecs
            预先生成的参数信息
        args
            位置参数
        kwargs
            关键字参数
        """
        self.cls = cls
        if _argspecs is None:
            self.argspecs = ArgSpecs.of(cls)
        else:
            self.argspecs = _argspecs

        self.argspecs.push_args(*args)
        self.argspecs.bind_kwargs(**kwargs)

    def find_param(self, predicate, remove=False, repl=None, default=undefined):
        """在参数列表中搜索指定参数

        Parameters
        ----------
        predicate
            搜索判断条件
        remove
            指示是否移除搜索到的参数
        repl
            替换项，如果搜索到结果且repl不为None则用该参数替换

        Returns
        -------
        ret
            搜索到的目标参数
        """
        ret = default
        matched = False

        for i, val in self.argspecs.bound_posargs.items():
            argspec = self.argspecs.posarg_specs[i]
            matched = predicate(i, val)

            if not matched:
                matched = matched or predicate(argspec.name, val)

            if matched:
                ret = val
                if repl is not None:
                    self.argspecs.bind_arg(argspec, repl)
                elif remove:
                    self.argspecs.remove(argspec)
                break

        if not matched:
            for argspec, val in self.argspecs.bound_kwargs.items():
                matched = predicate(argspec.name, val)

                if matched:
                    ret = val
                    if repl is not None:
                        self.argspecs.bind_arg(argspec, repl)
                    elif remove:
                        self.argspecs.remove(argspec)
                    break

        if ret == undefined:
            raise ValueError('Bound arg not found.')

        return ret

    def find_param_by_type(self, type, remove=False, repl=None):
        """在参数列表中搜索指定类型的参数

        Parameters
        ----------
        type
            指定的参数类型
        remove
            指示是否移除搜索到的参数
        repl
            替换项，如果搜索到结果且repl不为None则用该参数替换

        Returns
        -------
        ret
            搜索到的目标参数
        """
        try:
            return self.find_param(lambda k, v: isinstance(v, type), remove=remove, repl=repl)
        except ValueError:
            raise ValueError(f'Bound arg with type `{type}` not found.')

    def find_param_by_name(self, name, remove=False, repl=None):
        """在参数列表中搜索指定名字的参数

        Parameters
        ----------
        name
            指定的参数名
        remove
            指示是否移除搜索到的参数
        repl
            替换项，如果搜索到结果且repl不为None则用该参数替换

        Returns
        -------
        ret
            搜索到的目标参数
        """
        try:
            return self.find_param(lambda k, v: k == name, remove=remove, repl=repl)
        except ValueError:
            raise ValueError(f'Bound arg with name `{name}` not found.')

    def construct(self, replace_by_type: dict[type, Any] = None, **new_kwargs):
        """构造类对象

        Parameters
        ----------
        replace_by_type
            需要按类型替换的参数列表
        new_kwargs
            新的关键字参数

        Returns
        -------
        ret
            构造的类对象
        """
        temp_args = self.argspecs.clone()

        if replace_by_type is None:
            replace_by_type = {}

        if new_kwargs is None:
            new_kwargs = {}

        for argspec, val in itertools.chain(
                list(zip(temp_args.posarg_specs, temp_args.bound_posargs)),
                list(temp_args.bound_kwargs.items())
        ):
            name = argspec.name
            orig = val
            t = type(val)
            if t in replace_by_type:
                val = replace_by_type[t]
                t = type(val)

            if argspec.name in new_kwargs:
                val = new_kwargs.pop(name)
                t = type(val)

            if t == LazyClassFactory:
                val = val.construct()

            if val is not orig:
                temp_args.bind_arg(argspec, val)

        if len(new_kwargs) > 0:
            for k, v in new_kwargs.items():
                temp_args.bind_arg(k, v)

        return temp_args.call(self.cls)

    def clone(self):
        return LazyClassFactory(self.cls, _argspecs=self.argspecs.clone())

    def get(self, name_or_index, default=undefined):
        return self.argspecs.get_bound_arg(name_or_index, default=default)

    def pop(self, name_or_index, default=undefined):
        return self.argspecs.pop_bound_arg(name_or_index, default=default)

    def __getitem__(self, key):
        return self.argspecs.get_bound_arg(key)

    def __setitem__(self, key, value):
        return self.argspecs.bind_arg(key, value)

    def __delitem__(self, key):
        return self.argspecs.remove(key)



import numpy as np

from metalpy.utils.dhash import dhash
from .transform import Transform


class Translation(Transform):
    def __init__(self, x, y, z):
        super().__init__()
        self.delta = np.asarray([x, y, z])

    def transform(self, mesh):
        return np.asarray(mesh) + self.delta

    def inverse_transform(self, mesh):
        return np.asarray(mesh) - self.delta

    def clone(self):
        return Translation(*self.delta)

    def __dhash__(self):
        return dhash(*self.delta)



"""
坐标系为右手坐标系，东为+x，地理北为+y，竖直向上为+z
"SimPEG uses a right-handed coordinate system, with z being positive upwards"
"""
import json

import numpy as np
import pandas as pd
from matplotlib import pyplot as plt
from scipy.spatial.transform import Rotation as R

from metalpy.utils.misc import sin2pi, cos2pi, plot_opaque_cube, plot_linear_cube

from colour import Color


def inplaceable(func):
    def wrapper(*args, **kwargs):
        # 规避ide无法识别修饰器中的self参数
        self = args[0]
        args = args[1:]

        # 替换inplace行为
        obj = self.dispatch_inplace_delegate(kwargs.get('inplace'))
        return func(obj, *args, **kwargs)

    return wrapper


class ArraySlices:
    def __init__(self, arrays):
        self.arrays = arrays
        self.index = 0

    def slice(self, n_slices=1):
        ret = np.vstack(self.arrays[self.index: self.index + n_slices])
        self.index += n_slices
        return ret


class SensorArray:
    def __init__(self, pts, inplace=False):
        """传感器阵列

        Parameters
        ----------
        pts
            初始观测点集合
        inplace
            指示该实例的所有操作是否为inplace操作

        Notes
        -----
            Todo: 实现旋转平移等操作对1d和2d的支持
        """
        self.pts = np.asarray(pts, dtype=np.float64)
        self.inplace = inplace

        self.vt = np.zeros(3)
        self.vs = np.zeros(3)
        self.location_func = None
        self.orientation_func = None

    def __len__(self):
        return len(self.pts)

    def clear_movement(self):
        self.vt = np.zeros(3)
        self.vs = np.zeros(3)
        self.location_func = None
        self.orientation_func = None

    def set_default_inplace_behavior(self, b):
        """
        设置默认的inplace行为，True为所有操作均在当前类上修改，False为所有操作都构造新的实例
        :param b:
        :return:
        """
        self.inplace = b
        return self

    def clone(self, clear_movement=False):
        """
        构造一个新实例
        :return:
        """
        ret = SensorArray(self.pts.copy())
        ret.inplace = self.inplace

        if not clear_movement:
            ret.vt = self.vt
            ret.vs = self.vs
            ret.location_func = self.location_func
            ret.orientation_func = self.orientation_func

        return ret

    def dispatch_inplace_delegate(self, inplace):
        if inplace is None:
            inplace = self.inplace
        return self if inplace else self.clone()

    @inplaceable
    def scale(self, f, **kwargs):
        """
        警告，是关于原点放缩，可能不适用大部分情形
        :param f: 比例因子
        :return:
        """
        self.pts *= f
        return self

    @inplaceable
    def rotate(self, y, a, b, rot_tags='xyz', **kwargs):
        rot = R.from_euler(rot_tags, [y, a, b], degrees=True).as_matrix()
        self.pts = self.pts.dot(rot.T)
        return self

    @inplaceable
    def target_rotate(self, y, a, b, rot_tags='xyz', **kwargs):
        rot = R.from_euler(rot_tags, [-y, -a, -b], degrees=True).as_matrix()
        return SensorArray(self.pts.dot(rot.T))

    @inplaceable
    def translate(self, dx, dy, dz, **kwargs):
        translate_mat = np.asarray([[dx, dy, dz]]).repeat(len(self.pts), axis=0)
        self.pts += translate_mat
        return self

    def get_points(self, n_chunks=1, worker_id=0):
        return np.array_split(self.pts, n_chunks)[worker_id]

    def cut(self, n_chunks):
        return ArraySlices(np.array_split(self.pts, n_chunks))

    """
    with_*** 系列函数用于运动状态下观测点生成，采用lazy evaluation，在执行run后才会计算具体的观测点序列
    """

    @inplaceable
    def with_target_speed(self, vt, **kwargs):
        self.vt = vt
        return self

    @inplaceable
    def with_sensor_speed(self, vs, **kwargs):
        self.vs = vs
        return self

    @inplaceable
    def with_trajectory(self, location_func, **kwargs):
        """
        :param location_func:原点位置关于时间的函数, t -> (x, y, z)
        :return:
        """
        self.location_func = location_func
        return self

    @inplaceable
    def with_pose_movement(self, orientation_func, **kwargs):
        """
        :param orientation_func:朝向关于时间的函数, t -> (y, a, b)
        :return:
        """
        self.orientation_func = orientation_func
        return self

    def __default_get_location(self, t, **kwargs):
        relative_velocity = self.vs - self.vt
        delta = relative_velocity * t
        return delta

    def __default_get_orientation(self, t, **kwargs):
        return [0, 0, 0]

    def construct(self, dt=None, n_samples=None, timespan=None, with_timestamps=True):
        """
        参数三选二，采样结果包含0时刻的位置坐标（即初始位置坐标）
        :param dt:采样间隔
        :param n_samples:采样次数
        :param timespan:采样持续时间
        :param with_timestamps:是否返回每个观测点对应的时间刻，True则返回array, times，False则返回array
        :return:依据运动函数得到的一组点序列，作为一个新的SensorArray返回，新实例的运动函数会被重置为无运动
        """

        if dt is None:
            dt = timespan / (n_samples - 1)

        if timespan is None:
            timespan = dt * (n_samples - 1)

        ret = []
        timerange = np.arange(0, timespan + dt, dt)

        location_func = self.location_func
        orientation_func = self.orientation_func

        if location_func is None:
            def location_func(tx): return self.__default_get_location(tx)

        if orientation_func is None:
            def orientation_func(tx): return self.__default_get_orientation(tx)

        for t in timerange:
            delta = location_func(t)
            orientation = orientation_func(t)
            ret.append(self.rotate(*orientation, inplace=False).translate(*delta, inplace=False).get_points())
            # ret.append(self.pts + delta)

        pts = np.vstack(ret)
        ret = self.clone(clear_movement=True)
        ret.pts = pts

        if with_timestamps:
            return ret, timerange.repeat(len(self.pts))
        else:
            return ret

    def to_file(self, fname, **kwargs):
        with open(fname, 'w') as f:
            json.dump(self.pts.tolist(), f)

    def from_file(self, fname):
        with open(fname, 'r') as f:
            self.pts = np.asarray(json.load(f))

    def to_csv(self, fname):
        with open(fname, 'w') as output:
            pd.DataFrame(self.pts).to_csv(output, index=False, header=False, line_terminator='\n')

    @staticmethod
    def concat(s1, s2):
        return SensorArray(np.r_[s1.pts, s2.pts])

    @staticmethod
    def from_csv(fname):
        with open(fname, 'r') as _input:
            pts = pd.read_csv(_input, header=None).to_numpy()
        return SensorArray(pts)


def get_std7pts(distance):
    array = SensorArray([
        [0, 0, 0],
        [distance, 0, 0],
        [0, distance, 0],
        [0, 0, distance],
        [-distance, 0, 0],
        [0, -distance, 0],
        [0, 0, -distance],
    ])

    return array


def get_ext10pts(distance):
    array = SensorArray([
        [0, 0, 0],
        [distance, 0, 0],
        [0, distance, 0],
        [0, 0, distance],
        [-distance, 0, 0],
        [0, -distance, 0],
        [0, 0, -distance],
        [distance, distance, 0],
        [0, distance, distance],
        [distance, 0, distance],
    ])

    return array


def get_grids(distance, nx, ny, nz):
    xs, ys, zs = np.meshgrid(np.linspace(0, (nx - 1) * distance, nx), np.linspace(0, (ny - 1) * distance, ny),
                             np.linspace(0, (nz - 1) * distance, nz))
    xs, ys, zs = xs.flatten(), ys.flatten(), zs.flatten()
    array = SensorArray(np.c_[xs, ys, zs])

    return array


def _check_single_and_iterable(value, value_array, index):
    if value is not None:
        return value
    if value_array is not None:
        if hasattr(value_array, '__getitem__'):
            if len(value_array) > index:
                return value_array[index]
            else:
                return None
        else:
            return value_array
    return None


def _is_all_not_none(*args):
    for arg in args:
        if arg is None:
            return False
    return True


def _get_seq(dx, nx, xmin, xmax, lx, xs):
    if xs is not None:
        return xs

    while True:
        if lx is None:
            if nx is not None:
                if nx == 1:
                    lx = 0
                elif dx is not None:
                    lx = dx * (nx - 1)

        # nx, xmin, xmax
        if xmin is None:
            if _is_all_not_none(xmax, lx):
                xmin = xmax - lx
            else:
                break

        if xmax is None:
            if _is_all_not_none(xmin, lx):
                xmax = xmin + lx
            else:
                break

        if lx is None:
            lx = xmax - xmin

        if nx is None:
            if dx is None:
                break
            nx = np.ceil(lx / dx).astype(int) + 1

        xs = np.linspace(xmin, xmax, nx)
        return xs

    return None


def get_grids_ex(
        dx=None, dy=None, dz=None, cell_width=None,
        nx=None, ny=None, nz=None, n=None,
        xmin=None, ymin=None, zmin=None, origin=None,
        xmax=None, ymax=None, zmax=None, end=None,
        lx=None, ly=None, lz=None, length=None,
        xs=None, ys=None, zs=None,
):
    """通过给定序列参数获取观测网格

    Parameters
    ----------
    dx, dy, dz, cell_width
        x, y, z方向网格大小
    nx, ny, nz, n
        x, y, z方向网格数，包含起点和终点
    xmin, ymin, zmin, origin
        x, y, z方向网格起点
    xmax, ymax, zmax, end
        x, y, z方向网格重点
    lx, ly, lz, length
        x, y, z方向网格长度
    xs, ys, zs
        x, y, z方向序列

    Returns
    -------
    ret
        指定规格下的SensorArray
    """
    dx = _check_single_and_iterable(dx, cell_width, 0)
    nx = _check_single_and_iterable(nx, n, 0)
    xmin = _check_single_and_iterable(xmin, origin, 0)
    xmax = _check_single_and_iterable(xmax, end, 0)
    lx = _check_single_and_iterable(lx, length, 0)
    xs = _get_seq(dx, nx, xmin, xmax, lx, xs)

    dy = _check_single_and_iterable(dy, cell_width, 1)
    ny = _check_single_and_iterable(ny, n, 1)
    ymin = _check_single_and_iterable(ymin, origin, 1)
    ymax = _check_single_and_iterable(ymax, end, 1)
    ly = _check_single_and_iterable(ly, length, 1)
    ys = _get_seq(dy, ny, ymin, ymax, ly, ys)

    dz = _check_single_and_iterable(dz, cell_width, 2)
    nz = _check_single_and_iterable(nz, n, 2)
    zmin = _check_single_and_iterable(zmin, origin, 2)
    zmax = _check_single_and_iterable(zmax, end, 2)
    lz = _check_single_and_iterable(lz, length, 2)
    zs = _get_seq(dz, nz, zmin, zmax, lz, zs)

    seqs = [xs]
    if ys is not None:
        seqs.append(ys)
        if zs is not None:
            seqs.append(zs)

    seqs = np.meshgrid(*seqs, indexing='ij')
    seqs = tuple(seq.ravel() for seq in seqs)
    pts = np.c_[seqs]
    return SensorArray(pts)


if __name__ == '__main__':

    # 构造网格，间隔10m，xyz方向各行
    array = get_grids(100, 2, 2, 2)

    # 设置array的各个操作默认是否为inplace，即是否会修改自身
    # array.inplace = False

    fig = plt.figure(figsize=(9, 4))

    ax = plt.axes(projection='3d')
    ax.set_zlim(0, 800)

    ax.set_xlabel('X(m)')
    ax.set_ylabel('Y(m)')
    ax.set_zlabel('Z(m)')
    b = 512
    plot_opaque_cube(ax, -b, -b, 30, dx=2 * b, dy=2 * b, dz=1, alpha=0.4)
    plot_opaque_cube(ax, -b, -b, 300, dx=2 * b, dy=2 * b, dz=1, alpha=0.4)
    plot_opaque_cube(ax, -b, -b, 600, dx=2 * b, dy=2 * b, dz=1, alpha=0.4)

    a, b, c = 80.0, 10.0, 10.0
    u = np.linspace(0, 2 * np.pi, 100)
    v = np.linspace(0, np.pi, 100)
    x = a * np.outer(np.cos(u), np.sin(v))
    y = b * np.outer(np.sin(u), np.sin(v))
    z = c * np.outer(np.ones(np.size(u)), np.cos(v))

    # Plot the surface
    ax.plot_surface(x, y, z, color='b')

    plt.show()

    ax = plt.axes(projection='3d')
    ax.set_xlim(-5, 5)
    ax.set_ylim(-5, 5)
    ax.set_zlim(-5, 5)
    plot_linear_cube(ax, 0, 0, 0, dx=3.33333, dy=1, dz=1)

    plt.show()

    # 指定目标速度或阵列速度
    linear_movement, ts = array.with_target_speed(np.array([1, 2, 3])).construct(dt=0.2, timespan=5, with_timestamps=True)

    ax = plt.axes(projection='3d')

    red = Color("red")
    colors = list(red.range_to(Color("blue"), len(linear_movement)))
    x, y, z = linear_movement.get_points().T
    ax.scatter(*array.get_points().T)
    # ax.scatter(x, y, z, c=ts, cmap='plasma')
    # ax.quiver(850, 815, -785, -200, 0, 0)
    ax.quiver(50, 50, 50, 200, 0, 0)

    plot_linear_cube(ax, 800, 800, -800, dx=100, dy=30, dz=30)

    plt.show()

    # 指定目标运动轨迹
    circular_movement, ts = get_grids(0, 1, 1, 1).with_target_speed(np.array([20, 0, 0])).with_trajectory(
        location_func=lambda t: 100*np.array([sin2pi(t / 5), cos2pi(t / 5), 0]) - np.array([20,0,0]) * t).construct(dt=0.2, timespan=5,
                                                                                      with_timestamps=True)

    circular_movement = circular_movement.translate(800, 800, 0)
    ax = plt.axes(projection='3d')
    pts = circular_movement.get_points()
    ax.scatter(*pts.T, c=ts, cmap='plasma')

    plot_linear_cube(ax, 800, 800, -800, dx=100, dy=30, dz=30)
    ax.quiver(850, 815, -785, -200, 0, 0)

    # ax.quiver(*pts[-2], *(pts[-1] - pts[-2]), arrow_length_ratio=0.0001)
    plt.show()

    # 通过运动生成初始点阵然后再进行新的运动
    compound_movement, ts = get_grids(0, 1, 1, 1) \
        .with_sensor_speed(np.array([1, 1, 1])) \
        .construct(n_samples=10, timespan=5, with_timestamps=False) \
        .with_trajectory(location_func=lambda t: np.array([sin2pi(t / 5), cos2pi(t / 5), 0])) \
        .construct(dt=0.2, timespan=5, with_timestamps=True)

    ax = plt.axes(projection='3d')
    ax.scatter(*compound_movement.get_points().T, c=ts, cmap='plasma')
    plt.show()



from __future__ import annotations

from typing import overload, Collection

import numpy as np
from numpy.typing import NDArray
from scipy import stats

from metalpy.aero.utils.line_analysis import check_slope
from metalpy.utils.array_ops import get_column, cstack, get_row
from . import AerialSurvey, AerialSurveyPoint


class AerialSurveyLine(AerialSurvey):
    def __init__(self, position, data=None):
        super().__init__(position, data)

    def get_line_spec(self, robust=True, check_std=True):
        x, y = np.asarray(self.position).T

        if check_std:
            rx, ry = check_slope(x, y)
        else:
            rx, ry = x, y

        if robust:
            slope, intercept = stats.siegelslopes(ry, rx)
        else:
            slope, intercept, r_value, p_value, std_err = stats.linregress(rx, ry)

        if rx is x:
            return slope, -1, intercept
        else:
            return -1, slope, intercept

    @overload
    def align(self, points: NDArray, *, robust: bool = True, check_std: bool = True) -> np.ndarray: ...

    @overload
    def align(self, *, robust: bool = True, check_std: bool = True) -> AerialSurveyLine: ...

    def align(self, points=None, *, robust=True, check_std=True):
        if points is not None:
            position = np.asarray(points)
        else:
            position = self.position

        x = get_column(position, 0)
        y = get_column(position, 1)
        a, b, c = self.get_line_spec(robust=robust, check_std=check_std)
        if b == -1:
            y = a * x + c
        else:
            x = b * y + c

        if points is None:
            return AerialSurveyLine(cstack(x, y), self.data)
        else:
            return cstack(x, y)

    def __len__(self):
        return len(self.position)

    @overload
    def __getitem__(self, item: int) -> AerialSurveyPoint: ...

    @overload
    def __getitem__(self, item: slice | Collection[int]) -> AerialSurveyLine: ...

    def __getitem__(self, item):
        """对测线进行切片

        传入单个值索引时会返回测点对象

        默认按行号取值，如果需要取pandas的行索引，请使用 `pd.Index` 进行包装

        Parameters
        ----------
        item
            索引

        Returns
        -------
        sliced_line
            索引切片得到的新测线或测点
        """
        if self.data is not None:
            data = get_row(self.data, item)
        else:
            data = None

        position = get_row(self.position, item)

        if isinstance(item, int):
            return AerialSurveyPoint(position, data)
        else:
            return AerialSurveyLine(position, data)



from __future__ import annotations

import warnings
from typing import Iterable, Callable


class Dummy:
    def __init__(self, name):
        self.name = name

    def __hash__(self):
        return hash(self.name)

    def __eq__(self, other):
        return isinstance(other, Dummy) and self.name == other.name


undefined = Dummy('__undefined')


def stringify(e):
    return [str(i) for i in e]


def ensure_as_iterable(arg, excludes=None):
    if excludes is None:
        excludes = []
    elif not isinstance(excludes, Iterable):
        excludes = [excludes]

    if arg is None:
        return []
    for exclude in excludes:
        if isinstance(arg, exclude):
            return [arg]
    if not isinstance(arg, Iterable):
        return [arg]

    return arg


def get_or_default(dictionary, key, _default=None, default=None, supplier: Callable | None = None, remove=False):
    """获取键对应的值，若不存在则返回指定的默认值

    Parameters
    ----------
    dictionary
        映射
    key
        键
    _default
        默认值

        .. deprecated::
            使用default
    default
        默认值
    supplier
        默认值提供器，需要返回默认值且未指定default时，将尝试调用supplier获取返回值
    remove
        指示找到键值对时是否删除该元素

    Returns
    -------
    ret
        dictionary[key]，若`key`不存在且`default`非空则返回`default`，否则调用`supplier()`作为返回值
    """
    if key in dictionary:
        ret = dictionary[key]
        if remove:
            del dictionary[key]
    else:
        if default is None:
            default = _default
        if default is None and supplier is not None:
            default = supplier()
        ret = default

    return ret


def pop_or_default(dictionary, key, _default=None, default=None, supplier: Callable | None = None):
    """获取并移除键对应的值，若不存在则返回指定的默认值

    Parameters
    ----------
    dictionary
        映射
    key
        键
    _default
        默认值

        .. deprecated::
            使用default
    default
        默认值
    supplier
        默认值提供器，需要返回默认值且未指定default时，将尝试调用supplier获取返回值

    Returns
    -------
    ret
        dictionary[key]，若`key`不存在且`default`非空则返回`default`，否则调用`supplier()`作为返回值

    See Also
    --------
    get_or_default
    """
    return get_or_default(
        dictionary=dictionary,
        key=key,
        _default=_default,
        default=default,
        supplier=supplier,
        remove=True
    )


def not_none_or_default(val, _default=None, default=None, supplier: Callable | None = None):
    """判断指定值是否为空，非空则返回原值，否则返回指定的默认值

    Parameters
    ----------
    val
        待判断的值
    _default
        默认值

        .. deprecated::
            使用default
    default
        默认值
    supplier
        默认值提供器，需要返回默认值且未指定default时，将尝试调用supplier获取返回值

    Returns
    -------
    ret
        若`val`非空则返回`val`，若`val`为None且`default`非空则返回`default`，否则调用`supplier`作为返回值
    """
    if val is None:
        if default is None:
            default = _default
        if default is None and supplier is not None:
            default = supplier()
        return default
    else:
        return val


def get_first_key(dictionary):
    for k in dictionary:
        return k

    return None


def ensure_set_key(dictionary, key, value, *, transfer=None):
    while key in dictionary:
        if transfer is None:
            def transfer(x):
                return f'_{x}_'
        key = transfer(key)

    dictionary[key] = value
    return key


def get_params_dict(**kwargs):
    return dict(kwargs)


def get_params(*args, **kwargs):
    return list(args), dict(kwargs)


def is_numeric_array(obj):
    dtype = getattr(obj, 'dtype', None)
    if dtype is None:
        return False

    # 用于进一步判断numpy和pandas的元素类型
    # torch数组只支持数值类型，因此不需要继续判断，用空字符串跳过后续分支
    dtype_name = getattr(dtype, 'name', '')

    non_numeric_keywords = [
        'str',  # numpy - 'str', pandas - 'string'
        'object'  # numpy, pandas - 'object'
    ]
    for k in non_numeric_keywords:
        if k in dtype_name:
            return False

    return True


def notify_package(pkg_name, reason, install=None):
    if install is None:
        install = f'pip install {pkg_name}'

    warnings.warn(
        f'{reason}'
        f'\nConsider install `{pkg_name}` with following command:'
        f'\n'
        f'\n  >>> {install}'
        f'\n'
    )



from numpy.testing import assert_equal

from metalpy.scab.modelling import Scene
from metalpy.scab.modelling.shapes import Cuboid


def create_mesh():
    """生成一个长宽高为3个网格的27个网格组成的立方体，
    中心的1个网格被设置为不同的类型，
    构建网格时网格边界向外延申1格（即外部一圈为无效网格）
    """
    scene = Scene()
    cell_size = 1
    cube_cells = 3

    outer_cube = Cuboid(corner=[0, 0, 0], size=cube_cells * cell_size)
    scene.append(outer_cube, models={'value': 1, 'type': 0})
    scene.append(Cuboid(center=outer_cube.center, size=cell_size), models={'value': 2, 'type': 1})
    bounds = scene.bounds

    model_mesh = scene.build(cell_size=cell_size, bounds=bounds.expand(increment=cell_size))
    model_mesh.set_active_model('value')

    return model_mesh


def test_extract():
    model_mesh = create_mesh()
    model_mesh = model_mesh.extract(model_mesh['type'] == 1)
    mask = model_mesh.check_complete_mask([model_mesh.n_cells // 2])  # 即正中间的网格
    assert_equal(model_mesh.active_cells, mask)


def test_reactivate():
    model_mesh = create_mesh()
    mesh = model_mesh.mesh
    other_model_mesh = model_mesh.reactivate(
        (mesh.cell_centers - mesh.cell_centers[0]).sum(axis=1) <= 6
    )

    import numpy as np
    assert np.count_nonzero(other_model_mesh['value'] == 1) == 16
    assert np.count_nonzero(other_model_mesh['value'] == 2) == 1


def test_rebind():
    model_mesh = create_mesh()
    other_model_mesh = model_mesh.rebind('type')
    assert other_model_mesh.n_active_cells == 1



import re

import pytest

from metalpy.utils.regex_pattern import expand_regex_pattern
from metalpy.utils.string import format_string_list


def check_expansion(pat, assert_size=None, limit=None):
    try:
        re_pat = re.compile(r'(?!\\)\.').sub(r'\.', pat)
        pattern = re.compile(re_pat)
    except re.error as _:
        with pytest.raises(SyntaxError) as exc_info:
            exc_info.message = f'Inconsistent behavior between `re.compile` and `RegexPattern` for `{pat}`.'
            for _ in expand_regex_pattern(pat, limit=1):
                pass
        return

    if assert_size is not None:
        # 检查展开后的字符串数量是否符合预期
        limit = assert_size + 2
        expansion = expand_regex_pattern(pat, limit=limit)
        expansion = tuple(expansion)

        if len(expansion) >= limit:
            at_least = 'at least '
        else:
            at_least = ''

        # 看起来pytest只会对非简单 assert 输出额外信息
        satisfied = len(expansion) == assert_size
        assert satisfied, (
            f'Unexpected expansion count (got {at_least}{len(expansion)}, expected {assert_size})'
            f' for `{pat}`. Got:'
            f'{format_string_list(expansion, multiline=True, limit=assert_size + 1)}'
        )
    else:
        expansion = expand_regex_pattern(pat, limit=limit)
        if limit is not None:
            # 设定了limit，顺便检查一下limit是否有效
            expansion = tuple(expansion)
            assert len(expansion) <= limit, f'Failed to set `limit` for `{pat}`.'

    for result in expansion:
        assert pattern.fullmatch(result), (
            f'Inconsistent result between `re.compile` and `RegexPattern` for `{pat}`.'
            f' Got `{result}`.'
        )


def test_regex_pattern_range():
    check_expansion(r'\[d-b]', assert_size=1)  # 转义范围符号
    check_expansion(r'[d-b]')  # 错误的范围
    check_expansion(r'[\b-b]')  # 空格到b
    check_expansion(r'[b-\b]')  # 错误的范围
    check_expansion(r'[b-\d]')  # 错误的范围
    check_expansion(r'[\\\-\d-b]')  # 范围内转义
    check_expansion(r'[\db]')  # 转义范围


def test_regex_pattern_repeat():
    check_expansion(r'(a|b)+', limit=10)
    check_expansion(r'(a|b)*', limit=3)
    check_expansion(r'(a|b)?', limit=3)
    check_expansion(r'a{0,3}', assert_size=4)
    check_expansion(r'a{,3}', assert_size=4)
    check_expansion(r'a{3,}', limit=3)


def test_regex_pattern_comprehensive():
    check_expansion('test[1234].(com|cn)/[012]{0,3}', limit=20)

    pat = r'https://webrd0[1-4].is.autonavi.com/appmaptile\?x=\{x}&y=\{y}&z=\{z}'
    check_expansion(pat, assert_size=4)



from .extends import extends
from .replaces import replaces
from .after import after
from .before import before
from .recoverable_injector import RecoverableInjector
from .function_context import terminate, terminate_with, modify_params
from .replacement import reverted, is_or_is_replacement



from __future__ import annotations

from enum import IntFlag
import functools

from metalpy.mexin import Mixin


class Profiler(IntFlag):
    Disabled = 0     # 无Profiler
    Scoped = 1       # 输出Taichi处理编译等中间过程的耗时情况
    CountKernel = 2  # 输出Taichi Kernel的运行情况，但相同kernel合并
    TraceKernel = 4  # 输出所有Taichi Kernel的运行情况

    Default = Scoped | CountKernel

    @property
    def needs_kernel_profiler(self):
        return self > 1

    @staticmethod
    def of(val):
        if val is True:
            return Profiler.Default
        elif val is False:
            return Profiler.Disabled
        else:
            return val


class TiedMixin(Mixin):
    def __init__(self, this, profile: Profiler | bool):
        super().__init__(this)

        self.profile = Profiler.of(profile)


def tied_profile(fn):
    """用于包装需要对TaichiKernel进行profile的成员函数，成员函数所在类需要包含profile属性
    （继承自TaichiKernelMixin自动包含该属性）

    kernel_profiler需要在ti.init时指定`kernel_profiler=True`
    """
    @functools.wraps(fn)
    def wrapper(self: TiedMixin, *args, **kwargs):
        import taichi as ti

        ret = fn(self, *args, **kwargs)
        profile = self.profile

        if profile & Profiler.Scoped:
            ti.profiler.print_scoped_profiler_info()

        if profile & Profiler.CountKernel:
            ti.profiler.print_kernel_profiler_info('count')

        if profile & Profiler.TraceKernel:
            ti.profiler.print_kernel_profiler_info('trace')

        if profile.needs_kernel_profiler:
            ti.profiler.clear_kernel_profiler_info()

        return ret

    return wrapper



from metalpy.utils.dhash import dhash
from .transform import Transform


class CompositeTransform(Transform):
    def __init__(self):
        super().__init__()
        self.transforms: list[Transform] = []

    def add(self, transform):
        self.transforms.append(transform)

    def transform(self, mesh):
        for trans in self.transforms:
            mesh = trans.transform(mesh)
        return mesh

    def inverse_transform(self, mesh):
        for trans in reversed(self.transforms):
            mesh = trans.inverse_transform(mesh)
        return mesh

    def clone(self):
        ret = CompositeTransform()
        for trans in self.transforms:
            ret.add(trans)

        return ret

    def __dhash__(self):
        return dhash(*self.transforms,)
